"""This module defines an ASE interface to NWchem

https://nwchemgit.github.io
"""
import os
from pathlib import Path

import numpy as np

from ase import io
from ase.calculators.calculator import FileIOCalculator
from ase.spectrum.band_structure import BandStructure
from ase.units import Hartree


class NWChem(FileIOCalculator):
    implemented_properties = ['energy', 'free_energy',
                              'forces', 'stress', 'dipole']
    _legacy_default_command = 'nwchem PREFIX.nwi > PREFIX.nwo'
    accepts_bandpath_keyword = True
    discard_results_on_any_change = True

    fileio_rules = FileIOCalculator.ruleset(
        extend_argv=['{prefix}.nwi'],
        stdout_name='{prefix}.nwo')

    def __init__(self, restart=None,
                 ignore_bad_restart_file=FileIOCalculator._deprecated,
                 label='nwchem', atoms=None, command=None, **kwargs):
        """
        NWChem keywords are specified using (potentially nested)
        dictionaries. Consider the following input file block::

            dft
                odft
                mult 2
                convergence energy 1e-9 density 1e-7 gradient 5e-6
            end

        This can be generated by the NWChem calculator by using the
        following settings:
        >>> from ase.calculators.nwchem import NWChem
        >>> calc = NWChem(dft={'odft': None,
        ...                    'mult': 2,
        ...                    'convergence': {'energy': 1e-9,
        ...                                    'density': 1e-7,
        ...                                    'gradient': 5e-6,
        ...                                    },
        ...                    },
        ...               )

        In addition, the calculator supports several special keywords:

        theory: str
            Which NWChem module should be used to calculate the
            energies and forces. Supported values are ``'dft'``,
            ``'scf'``, ``'mp2'``, ``'ccsd'``, ``'tce'``, ``'tddft'``,
            ``'pspw'``, ``'band'``, and ``'paw'``. If not provided, the
            calculator will attempt to guess which theory to use based
            on the keywords provided by the user.
        xc: str
            The exchange-correlation functional to use. Only relevant
            for DFT calculations.
        task: str
            What type of calculation is to be performed, e.g.
            ``'energy'``, ``'gradient'``, ``'optimize'``, etc. When
            using ``'SocketIOCalculator'``, ``task`` should be set
            to ``'optimize'``. In most other circumstances, ``task``
            should not be set manually.
        basis: str or dict
            Which basis set to use for gaussian-type orbital
            calculations. Set to a string to use the same basis for all
            elements. To use a different basis for different elements,
            provide a dict of the form:

            >>> calc = NWChem(...,
            ...               basis={'O': '3-21G',
            ...                      'Si': '6-31g'})

        basispar: str
            Additional keywords to put in the NWChem ``basis`` block,
            e.g. ``'rel'`` for relativistic bases.
        symmetry: int or str
            The point group (for gaussian-type orbital calculations) or
            space group (for plane-wave calculations) of the system.
            Supports both group names (e.g. ``'c2v'``, ``'Fm3m'``) and
            numbers (e.g. ``225``).
        autosym: bool
            Whether NWChem should automatically determine the symmetry
            of the structure (defaults to ``False``).
        center: bool
            Whether NWChem should automatically center the structure
            (defaults to ``False``). Enable at your own risk.
        autoz: bool
            Whether NWChem should automatically construct a Z-matrix
            for your molecular system (defaults to ``False``).
        geompar: str
            Additional keywords to put in the NWChem `geometry` block,
            e.g. ``'nucleus finite'`` for gaussian-shaped nuclear
            charges. Do not set ``'autosym'``, ``'center'``, or
            ``'autoz'`` in this way; instead, use the appropriate
            keyword described above for these settings.
        set: dict
            Used to manually create or modify entries in the NWChem
            rtdb. For example, the following settings enable
            pseudopotential filtering for plane-wave calculations::

                set nwpw:kbpp_ray .true.
                set nwpw:kbpp_filter .true.

            These settings are generated by the NWChem calculator by
            passing the arguments:

            >>> calc = NWChem(...,
            >>>               set={'nwpw:kbpp_ray': True,
            >>>                    'nwpw:kbpp_filter': True})

        kpts: (int, int, int), or dict
            Indicates which k-point mesh to use. Supported syntax is
            similar to that of GPAW. Implies ``theory='band'``.
        bandpath: BandPath object
            The band path to use for a band structure calculation.
            Implies ``theory='band'``.
        pretasks: list of dict
            Tasks used to produce a better initial guess
            for the wavefunction.
            These task typically use a cheaper level of theory
            or smaller basis set (but not both).
            The output energy and forces should remain unchanged
            regardless of the number of tasks or their parameters,
            but the runtime may be significantly improved.

            For example, a MP2 calculation preceded by guesses at the
            DFT and HF levels would be

            >>> calc = NWChem(theory='mp2', basis='aug-cc-pvdz',
            >>>               pretasks=[
            >>>                   {'dft': {'xc': 'hfexch'},
            >>>                    'set': {'lindep:n_dep': 0}},
            >>>                   {'theory': 'scf', 'set': {'lindep:n_dep': 0}},
            >>>               ])

            Each dictionary could contain any of the other parameters,
            except those which pertain to global configurations
            (e.g., geometry details, scratch dir).

            The default basis set is that of the final step in the calculation,
            or that of the previous step that which defines a basis set.
            For example, all steps in the example will use aug-cc-pvdz
            because the last step is the only one which defines a basis.

            Steps which change basis set must use the same theory.
            The following specification would perform SCF using the 3-21G
            basis set first, then B3LYP//3-21g, and then B3LYP//6-31G(2df,p)

            >>> calc = NWChem(theory='dft', xc='b3lyp', basis='6-31g(2df,p)',
            >>>               pretasks=[
            >>>                   {'theory': 'scf', 'basis': '3-21g',
            >>>                    'set': {'lindep:n_dep': 0}},
            >>>                   {'dft': {'xc': 'b3lyp'}},
            >>>               ])

            The :code:`'set': {'lindep:n_dep': 0}` option is highly suggested
            as a way to avoid errors relating to symmetry changes between tasks.

            The calculator will configure appropriate options for saving
            and loading intermediate wavefunctions, and
            place an "ignore" task directive between each step so that
            convergence errors in intermediate steps do not halt execution.
        """
        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file,
                                  label, atoms, command, **kwargs)
        if self.prefix is None:
            self.prefix = 'nwchem'
        self.calc = None

    def input_filename(self):
        return f'{self.prefix}.nwi'

    def output_filename(self):
        return f'{self.prefix}.nwo'

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        # Prepare perm and scratch directories
        perm = os.path.abspath(self.parameters.get('perm', self.label))
        scratch = os.path.abspath(self.parameters.get('scratch', self.label))
        os.makedirs(perm, exist_ok=True)
        os.makedirs(scratch, exist_ok=True)

        io.write(Path(self.directory) / self.input_filename(),
                 atoms, properties=properties,
                 label=self.label, **self.parameters)

    def read_results(self):
        output = io.read(Path(self.directory) / self.output_filename())
        self.calc = output.calc
        self.results = output.calc.results

    def band_structure(self):
        self.calculate()
        perm = self.parameters.get('perm', self.label)
        if self.calc.get_spin_polarized():
            alpha = np.loadtxt(os.path.join(perm, self.label + '.alpha_band'))
            beta = np.loadtxt(os.path.join(perm, self.label + '.beta_band'))
            energies = np.array([alpha[:, 1:], beta[:, 1:]]) * Hartree
        else:
            data = np.loadtxt(os.path.join(perm,
                                           self.label + '.restricted_band'))
            energies = data[np.newaxis, :, 1:] * Hartree
        eref = self.calc.get_fermi_level()
        if eref is None:
            eref = 0.
        return BandStructure(self.parameters.bandpath, energies, eref)



"""
Knowledgebase of Interatomic Models (KIM) Calculator for ASE written by:

Ellad B. Tadmor
Mingjian Wen
Daniel S. Karls
University of Minnesota

This calculator functions as a wrapper that selects an appropriate
calculator for a given KIM model depending on whether it supports the
KIM application programming interface (API) or not. For more information
on KIM, visit https://openkim.org.
"""

from . import kimpy_wrappers
from .calculators import (
    ASAPCalculator,
    KIMCalculator,
    LAMMPSLibCalculator,
    LAMMPSRunCalculator,
)
from .exceptions import KIMCalculatorError


def KIM(model_name, simulator=None, options=None, debug=False):
    """Calculator wrapper for OpenKIM models

    Returns a suitable calculator that can be used with any model
    archived in the Open Knowledgebase of Interatomic Models (OpenKIM)
    at https://openkim.org.  There are two kinds of models in KIM:
    Portable Models (PMs), which can be used with any KIM API-compliant
    simulator, and Simulator Models (SMs), which are essentially just
    wrappers around native commands in a specific simulator (often
    combined with values for the model parameters).  PMs published on
    openkim.org contain the string '__MO_' in their name, while SMs
    published on openkim.org contain the string '__SM_' in their name.

    Parameters
    ----------
    model_name : str
        The name of the KIM model installed on your system.  KIM models
        published on openkim.org follow a specific naming scheme (see
        https://openkim.org/doc/schema/kim-ids).

    simulator : str, optional
        Used to identify the ASE calculator that will be used.
        Currently supported values include 'kimmodel', 'lammpslib',
        'lammpsrun' and 'asap', and correspond to different calculators
        as follows:

        - kimmodel (default for PMs)
          : :py:mod:`ase.calculators.kim.kimmodel.KIMModelCalculator`

        - lammpsrun (PMs or LAMMPS SMs)
          : :py:mod:`ase.calculators.lammpsrun.LAMMPS`

        - lammpslib (default for LAMMPS SMs)
          : :py:mod:`ase.calculators.lammpslib.LAMMPSlib`

        - asap (PMs)
          : :py:mod:`asap3.Internal.OpenKIMcalculator.OpenKIMcalculator`

        - asap (ASAP SMs)
          : :py:mod:`asap3.Internal.BuiltinPotentials.EMT`

        In general, this argument should be omitted, in which case a
        calculator compatible with the specified model will
        automatically be determined.

    options : dict, optional
        Additional options passed to the initializer of the selected
        calculator.  If ``simulator`` == 'kimmodel', possible options are:

        - ase_neigh (bool)
          : Whether to use the kimpy neighbor list library (False) or
          use ASE's internal neighbor list mechanism (True). Usually
          kimpy's neighbor list library will be faster.  (Default:
          False)

        - neigh_skin_ratio (float)
          : The skin distance used for neighbor list construction,
          expressed as a fraction of the model cutoff (Default: 0.2)

        - release_GIL (bool)
          : Whether to release python GIL.  Releasing the GIL allows a KIM
          model to run with multiple concurrent threads. (Default: False)

        See the ASE LAMMPS calculators doc page
        (https://wiki.fysik.dtu.dk/ase/ase/calculators/lammps.html) for
        available options for the lammpslib and lammpsrun calculators.

    debug : bool, optional
        If True, detailed information is printed to stdout.  If the
        lammpsrun calculator is being used, this also serves as the
        value of the ``keep_tmp_files`` option. (Default: False)

    Returns
    -------
    ase.calculators.calculator.Calculator
        An ASE-compatible calculator.  Currently, this will be an instance of
        KIMModelCalculator, LAMMPS (the lammpsrun calculator), or LAMMPSlib,
        which are all defined in the ASE codebase, or an instance of either
        OpenKIMcalculator or EMT defined in the asap3 codebase.

    Raises
    ------
    KIMCalculatorError
        Indicates an error occurred in initializing the calculator,
        e.g. due to incompatible combinations of argument values
    """

    if options is None:
        options = {}

    # If this is a KIM Portable Model (supports KIM API), return
    # support through a KIM-compliant simulator
    model_type = "pm" if _is_portable_model(model_name) else "sm"

    if model_type == "pm":
        if simulator is None:  # Default
            simulator = "kimmodel"

        if simulator == "kimmodel":
            return KIMCalculator(model_name, options, debug)

        elif simulator == "asap":
            return ASAPCalculator(
                model_name, model_type, options=options, verbose=debug
            )

        elif simulator == "lammpsrun":
            supported_species = get_model_supported_species(model_name)

            # Return LAMMPS calculator
            return LAMMPSRunCalculator(
                model_name, model_type, supported_species, options, debug
            )

        elif simulator == "lammpslib":
            raise KIMCalculatorError(
                '"lammpslib" calculator does not support KIM Portable Models. '
                'Try using the "lammpsrun" calculator.'
            )
        else:
            raise KIMCalculatorError(
                'Unsupported simulator "{}" requested to run KIM '
                'Portable Model.'.format(simulator)
            )

    #######################################################
    # If we get to here, the model is a KIM Simulator Model
    #######################################################
    with kimpy_wrappers.SimulatorModel(model_name) as sm:

        # Handle default behavior for 'simulator'
        if simulator is None:
            if sm.simulator_name == "ASAP":
                simulator = "asap"
            elif sm.simulator_name == "LAMMPS":
                simulator = "lammpslib"

        if sm.simulator_name == "ASAP":

            return ASAPCalculator(
                model_name,
                model_type,
                options=options,
                model_defn=sm.model_defn,
                verbose=debug,
                supported_units=sm.supported_units,
            )

        elif sm.simulator_name == "LAMMPS":

            if simulator == "lammpsrun":

                return LAMMPSRunCalculator(
                    model_name,
                    model_type,
                    sm.supported_species,
                    options,
                    debug,
                    atom_style=sm.atom_style,
                    supported_units=sm.supported_units,
                )

            elif simulator == "lammpslib":
                return LAMMPSLibCalculator(
                    model_name, sm.supported_species, sm.supported_units,
                    options
                )

            else:
                raise KIMCalculatorError(
                    f'Unknown LAMMPS calculator: "{simulator}".'
                )

        else:
            raise KIMCalculatorError(
                f'Unsupported simulator: "{sm.simulator_name}".'
            )


def _is_portable_model(model_name):
    """
    Returns True if the model specified is a KIM Portable Model (if it
    is not, then it must be a KIM Simulator Model -- there are no other
    types of models in KIM)
    """
    with kimpy_wrappers.ModelCollections() as col:
        model_type = col.get_item_type(model_name)

    return (model_type ==
            kimpy_wrappers.wrappers.collection_item_type_portableModel)


def get_model_supported_species(model_name):
    if _is_portable_model(model_name):
        with kimpy_wrappers.PortableModel(model_name, debug=False) as pm:
            supported_species, _ = pm.get_model_supported_species_and_codes()
    else:
        with kimpy_wrappers.SimulatorModel(model_name) as sm:
            supported_species = sm.supported_species

    return supported_species



"""This test ensures that logging to a text file and to the trajectory file are
reporting the same values as in the ContourExploration object."""

import numpy as np
import pytest

from ase import io
from ase.md.contour_exploration import ContourExploration

from .test_ce_potentiostat import Al_block, bulk_Al_settings


def test_logging(testdir):

    seed = 19460926

    atoms = Al_block()
    atoms.rattle(stdev=0.18, seed=seed)

    rng = np.random.RandomState(seed)

    initial_energy = atoms.get_potential_energy()

    name = 'test_logging'

    traj_name = name + '.traj'
    log_name = name + '.log'

    with ContourExploration(
            atoms,
            **bulk_Al_settings,
            rng=rng,
            trajectory=traj_name,
            logfile=log_name,
    ) as dyn:
        energy_target = initial_energy
        dev = (atoms.get_potential_energy() - energy_target) / len(atoms)
        energy_targets = [energy_target]
        curvatures = [dyn.curvature]
        stepsizes = [dyn.step_size]
        deviation_per_atom = [dev]

        # we shift the target_energy to ensure it's actaully being
        # logged when it changes.
        de = 0.001 * len(atoms)

        # these print statements, mirror the log file.
        # print(energy_target, dyn.curvature, dyn.step_size, dev)

        for i in range(5):
            energy_target = initial_energy + de * i

            dyn.energy_target = energy_target
            dyn.run(1)
            dev = (atoms.get_potential_energy() - energy_target) / len(atoms)
            # print(energy_target, dyn.curvature, dyn.step_size, dev)

            energy_targets.append(energy_target)
            curvatures.append(dyn.curvature)
            stepsizes.append(dyn.step_size)
            deviation_per_atom.append(dev)

    # Now we check the contents of the log file
    # assert log file has correct length
    with open(log_name) as fd:
        length = len(fd.readlines())
    assert length == 7, length

    with io.Trajectory(traj_name, 'r') as traj, open(log_name) as fd:
        # skip the first line because it's a small initialization step
        lines = fd.readlines()[1:]
        for i, (im, line) in enumerate(zip(traj, lines)):

            lineparts = [float(part) for part in line.split()]

            log_energy_target = lineparts[1]
            assert 0 == pytest.approx(
                log_energy_target - energy_targets[i], abs=1e-5)

            log_energy = lineparts[2]
            assert 0 == pytest.approx(
                log_energy - im.get_potential_energy(), abs=1e-5)

            log_curvature = lineparts[3]
            assert 0 == pytest.approx(log_curvature - curvatures[i], abs=1e-5)

            log_step_size = lineparts[4]
            assert 0 == pytest.approx(log_step_size - stepsizes[i], abs=1e-5)

            log_dev = lineparts[5]
            assert 0 == pytest.approx(log_dev - deviation_per_atom[i], abs=1e-5)



"""Test file for exciting ASE calculator."""

import xml.etree.ElementTree as ET

import numpy as np
import pytest

import ase
import ase.calculators.exciting.exciting
import ase.calculators.exciting.runner

# Note this is an imitation of an exciting INFO.out output file.
# We've removed many of the lines of text that were originally in this outfile
# that are note usefule for testing purposes to save space in this file.
# We've also modified the file to contain a Ti atom and use an HCP cell to
# make the test more interesting since the HCP cell gives non-symmetric cell
# vectors in a cartesian basis set.

LDA_VWN_AR_INFO_OUT = """
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ Starting initialization                                                      +
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

 Lattice vectors (cartesian) :
     10.3360193975     10.3426010725      0.0054547264
    -10.3461511392     10.3527307290      0.0059928210
     10.3354645037     10.3540072605     20.6246241525

 Reciprocal lattice vectors (cartesian) :
      0.3039381122      0.3039214341     -0.3048853768
     -0.3036485697      0.3034554311     -0.0001760382
      0.0000078456     -0.0001685540      0.3047255253

 Unit cell volume                           :    4412.7512103067
 Brillouin zone volume                      :       0.0562121456

 Species :    1 (Ti)
     parameters loaded from                 :    Ti.xml
     name                                   :    titanium

     atomic positions (lattice) :
       1 :   0.00000000  0.00000000  0.00000000

 Total number of atoms per unit cell        :       1

 Spin treatment                             :    spin-unpolarised

 Number of Bravais lattice symmetries       :      48
 Number of crystal symmetries               :      48

 k-point grid                               :       1    1    1
 Total number of k-points                   :       1
 k-point set is reduced with crystal symmetries

 R^MT_min * |G+k|_max (rgkmax)              :      10.00000000
 Species with R^MT_min                      :       1 (Ti)
 Maximum |G+k| for APW functions            :       1.66666667
 Maximum |G| for potential and density      :       7.50000000
 Polynomial order for pseudochg. density    :       9

 G-vector grid sizes                        :      36    36    36
 Total number of G-vectors                  :   23871

 Maximum angular momentum used for
     APW functions                          :       8
     computing H and O matrix elements      :       4
     potential and density                  :       4
     inner part of muffin-tin               :       2

 Total nuclear charge                       :     -22.00000000
 Total electronic charge                    :      22.00000000
 Total core charge                          :      18.00000000
 Total valence charge                       :       4.00000000

 Effective Wigner radius, r_s               :       3.55062021

 Number of empty states                     :       5
 Total number of valence states             :      10

 Maximum Hamiltonian size                   :     263
 Maximum number of plane-waves              :     251
 Total number of local-orbitals             :      12

 Exchange-correlation type                  :     100
     libxc; exchange: Slater exchange; correlation: Vosko, Wilk & Nusair (VWN5)

 Smearing scheme                            :    Gaussian
 Smearing width                             :       0.00100000

 Using multisecant Broyden potential mixing

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ Ending initialization                                                        +
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ SCF iteration number :    1                                                  +
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Total energy                               :      -527.82493279
 _______________________________________________________________
 Fermi energy                               :        -0.20111449
 Kinetic energy                             :       530.56137212
 Coulomb energy                             :     -1029.02167746
 Exchange energy                            :       -27.93377198
 Correlation energy                         :        -1.43085548
 Sum of eigenvalues                         :      -305.07886015
 Effective potential energy                 :      -835.64023227
 Coulomb potential energy                   :      -796.81322609
 xc potential energy                        :       -38.82700618
 Hartree energy                             :       205.65681157
 Electron-nuclear energy                    :     -1208.12684923
 Nuclear-nuclear energy                     :       -26.55163980
 Madelung energy                            :      -630.61506441
 Core-electron kinetic energy               :         0.00000000

 DOS at Fermi energy (states/Ha/cell)       :         0.00000000

 Electron charges :
     core                                   :        10.00000000
     core leakage                           :         0.00000000
     valence                                :         8.00000000
     interstitial                           :         0.00183897
     charge in muffin-tin spheres :
                  atom     1    Ar          :        17.99816103
     total charge in muffin-tins            :        17.99816103
     total charge                           :        18.00000000

 Estimated fundamental gap                  :         0.36071248
        valence-band maximum at    1      0.0000  0.0000  0.0000
     conduction-band minimum at    1      0.0000  0.0000  0.0000

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
| Convergency criteria checked for the last 2 iterations                       +
| Convergence targets achieved. Performing final SCF iteration                 +
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 Total energy                               :      -527.81796101
 _______________________________________________________________
 Fermi energy                               :        -0.20044598
 Kinetic energy                             :       530.57303096
 Coulomb energy                             :     -1029.02642037
 Exchange energy                            :       -27.93372809
 Correlation energy                         :        -1.43084350
 Sum of eigenvalues                         :      -305.07413840
 Effective potential energy                 :      -835.64716936
 Coulomb potential energy                   :      -796.82023455
 xc potential energy                        :       -38.82693481
 Hartree energy                             :       205.65454603
 Electron-nuclear energy                    :     -1208.12932661
 Nuclear-nuclear energy                     :       -26.55163980
 Madelung energy                            :      -630.61630310
 Core-electron kinetic energy               :         0.00000000

 DOS at Fermi energy (states/Ha/cell)       :         0.00000000

 Electron charges :
     core                                   :        10.00000000
     core leakage                           :         0.00000000
     valence                                :         8.00000000
     interstitial                           :         0.00184037
     charge in muffin-tin spheres :
                  atom     1    Ar          :        17.99815963
     total charge in muffin-tins            :        17.99815963
     total charge                           :        18.00000000

 Estimated fundamental gap                  :         0.36095838
        valence-band maximum at    1      0.0000  0.0000  0.0000
     conduction-band minimum at    1      0.0000  0.0000  0.0000

++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ Self-consistent loop stopped                                                 +
| EXCITING NITROGEN-14 stopped                                                 =
"""


@pytest.fixture()
def nitrogen_trioxide_atoms():
    """Pytest fixture that creates ASE Atoms cell for other tests."""
    return ase.Atoms('NO3',
                     cell=[[2, 2, 0], [0, 4, 0], [0, 0, 6]],
                     scaled_positions=[(0, 0, 0), (0.25, 0.25, 0),
                                       (0, 0, 0.75), (0.5, 0.5, 0.5)],
                     pbc=True)


def test_ground_state_template_init(excitingtools):
    """Test initialization of the ExcitingGroundStateTemplate class."""
    gs_template_obj = (
        ase.calculators.exciting.exciting.ExcitingGroundStateTemplate())
    assert gs_template_obj.name == 'exciting'
    assert len(gs_template_obj.implemented_properties) == 2
    assert 'energy' in gs_template_obj.implemented_properties


def test_ground_state_template_write_input(
        tmp_path, nitrogen_trioxide_atoms, excitingtools):
    """Test the write input method of ExcitingGroundStateTemplate.

    We test is by writing a ground state calculation and a bandstructure
    calculation after that is run.

    Args:
        tmp_path: This tells pytest to create a temporary directory
             in which we will store the exciting input file.
        nitrogen_trioxide_atoms: pytest fixture to create ASE Atoms
            unit cell composed of NO3.
    """
    from excitingtools.input.bandstructure import (
        band_structure_input_from_ase_atoms_obj,
    )
    expected_path = tmp_path / 'input.xml'
    # Expected number of points in the bandstructure.
    expected_number_of_special_points = 12
    bandstructure_steps = 100
    binary_path = tmp_path / 'exciting_binary'

    gs_template_obj = (
        ase.calculators.exciting.exciting.ExcitingGroundStateTemplate())
    exciting_profile = ase.calculators.exciting.exciting.ExcitingProfile(
        command=str(binary_path))
    gs_template_obj.write_input(
        profile=exciting_profile,
        directory=tmp_path,
        atoms=nitrogen_trioxide_atoms,
        parameters={
            'title': None,
            'species_path': tmp_path,
            'ground_state_input': {
                'rgkmax': 8.0,
                'do': 'fromscratch',
                "ngridk": [6, 6, 6],
                'xctype': 'GGA_PBE_SOL',
                'vkloff': [0, 0, 0]},
            'properties_input': {
                'bandstructure': band_structure_input_from_ase_atoms_obj(
                    nitrogen_trioxide_atoms, steps=bandstructure_steps)}})
    # Let's assert the file we just wrote exists.
    assert expected_path.exists()
    # Let's assert it's what we expect.
    element_tree = ET.parse(expected_path)
    # Ensure the coordinates of the atoms in the unit cell is correct.
    # We could test the other parts of the input file related coming from
    # the ASE Atoms object like species data but this is tested already in
    # test/io/exciting/test_exciting.py.
    coords_list = element_tree.findall('./structure/species/atom')
    positions = np.array([[float(x)
                           for x in coords_list[i].get('coord').split()]
                          for i in range(len(coords_list))])
    assert positions == pytest.approx(
        nitrogen_trioxide_atoms.get_scaled_positions())

    # Ensure that the exciting calculator properites (e.g. functional type have
    # been set).
    assert element_tree.findall('input') is not None
    assert element_tree.getroot().tag == 'input'
    assert element_tree.getroot()[2].attrib['xctype'] == 'GGA_PBE_SOL'
    assert element_tree.getroot()[2].attrib['rgkmax'] == '8.0'
    # Ensure the bandstructure path is correct:
    band_path = element_tree.findall(
        './properties/bandstructure/plot1d/path')[0]
    assert band_path.tag == 'path'
    assert int(band_path.get('steps')) == bandstructure_steps
    assert len(list(band_path)) == expected_number_of_special_points


def test_ground_state_template_read_results(tmp_path, excitingtools):
    """Test the read result method of ExcitingGroundStateTemplate."""
    # ASE doesn't want us to store any other files for test, so instead
    # we copy an example exciting INFO.out file into the global variable
    # LDA_VWN_AR_INFO_OUT.
    output_file_path = tmp_path / 'info.xml'
    with open(output_file_path, "w", encoding="utf8") as xml_file:
        xml_file.write(LDA_VWN_AR_INFO_OUT)

    gs_template_obj = (
        ase.calculators.exciting.exciting.ExcitingGroundStateTemplate())
    results = gs_template_obj.read_results(tmp_path)
    final_scl_iteration = list(results["scl"].keys())[-1]
    assert pytest.approx(float(results["scl"][
        final_scl_iteration]["Hartree energy"])) == 205.65454603


def test_get_total_energy_and_bandgap(excitingtools):
    """Test getter methods for energy/bandgap results."""
    # Create a fake results dictionary that has two SCL cycles
    # and only contains values for the total energy and bandgap.
    results_dict = {
        'scl': {
            '1':
                {
                    'Total energy': '-240.3',
                    'Estimated fundamental gap': 2.0,
                },
            '2':
                {
                    'Total energy': '-242.3',
                    'Estimated fundamental gap': 3.1,
                }
        }

    }
    results_obj = ase.calculators.exciting.exciting.ExcitingGroundStateResults(
        results_dict)
    assert pytest.approx(results_obj.total_energy()) == -242.3
    assert pytest.approx(results_obj.band_gap()) == 3.1


def test_ground_state_calculator_init(tmpdir, excitingtools):
    """Test initiliazation of the ExcitingGroundStateCalculator"""
    ground_state_input_dict = {
        "rgkmax": 8.0,
        "do": "fromscratch",
        "ngridk": [6, 6, 6],
        "xctype": "GGA_PBE_SOL",
        "vkloff": [0, 0, 0]}
    calc_obj = ase.calculators.exciting.exciting.ExcitingGroundStateCalculator(
        runner=ase.calculators.exciting.runner.SimpleBinaryRunner(
            "exciting_serial", ['./'], 1, tmpdir, ['']),
        ground_state_input=ground_state_input_dict, directory=tmpdir)
    assert calc_obj.parameters["ground_state_input"]["rgkmax"] == 8.0



from ase import Atoms


def test_set_get_angle():
    "Test that set_angle() and get_angle() in Atoms are consistent"

    atoms = Atoms('HHCCHH', [[-1, 1, 0], [-1, -1, 0], [0, 0, 0],
                             [1, 0, 0], [2, 1, 0], [2, -1, 0]])

    list = [2, 3, 4]
    theta = 20
    old_angle = atoms.get_angle(*list)
    atoms.set_angle(*list, angle=old_angle + theta)
    new_angle = atoms.get_angle(*list)

    assert abs(new_angle - (old_angle + theta)) < 1.0e-9



import numpy as np

from ase.io.jsonio import decode, encode


def test_json_arrays():

    def check(obj):
        txt = encode(obj)
        newobj = decode(txt, always_array=False)
        print(obj, '-->', newobj)
        assert isinstance(obj, type(newobj)), '{} vs {}'.format(type(obj),
                                                                type(newobj))
        assert np.shape(obj) == np.shape(newobj)
        assert np.array_equal(obj, newobj)

    check([1, 2, 3])
    check([1.0, 2.0, 3.0])
    check([])

    check(np.arange(3))
    check(np.arange(3).astype(float))
    check(np.empty((3, 0, 7)))
    check(np.empty((0, 3, 7), dtype=int))
    check(np.ones(2, complex))
    check(np.ones(2, np.complex64))



from ase.build import add_adsorbate, fcc100
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.mep import NEB
from ase.optimize import QuasiNewton


def main():
    # 2x2-Al(001) surface with 3 layers and an
    # Au atom adsorbed in a hollow site:
    slab = fcc100('Al', size=(2, 2, 3))
    add_adsorbate(slab, 'Au', 1.7, 'hollow')
    slab.center(axis=2, vacuum=4.0)

    # Fix second and third layers:
    mask = [atom.tag > 1 for atom in slab]
    slab.set_constraint(FixAtoms(mask=mask))

    # Use EMT potential:
    slab.calc = EMT()

    # Initial state:
    qn = QuasiNewton(slab, logfile=None)
    qn.run(fmax=0.05)
    initial = slab.copy()

    # Final state:
    slab[-1].x += slab.get_cell()[0, 0] / 2
    qn = QuasiNewton(slab, logfile=None)
    qn.run(fmax=0.05)
    final = slab.copy()

    # Setup a NEB calculation
    constraint = FixAtoms(mask=[atom.tag > 1 for atom in initial])

    images = [initial]
    for _ in range(3):
        image = initial.copy()
        image.set_constraint(constraint)
        images.append(image)

    images.append(final)

    neb = NEB(images)
    neb.interpolate()

    for image in neb.images[1:-1]:
        image.calc = EMT()


if __name__ == '__main__':
    main()



from ase.build import bulk
from ase.calculators.calculator import compare_atoms
from ase.io import read, write


def test_convert(tmp_path, cli):
    infile = tmp_path / 'images.json'
    images = [bulk('Si'), bulk('Au')]
    write(infile, images, format='json')

    outfile = tmp_path / 'images.traj'
    cli.ase('convert', str(infile), str(outfile))
    images2 = read(outfile, ':')

    assert len(images2) == 2
    for a1, a2 in zip(images, images2):
        assert not compare_atoms(a1, a2)



import numpy as np

from ase.geometry import cell_to_cellpar as c2p
from ase.geometry import cellpar_to_cell as p2c

eps = 2 * np.spacing(90., dtype=np.float64)


def nearly_equal(a, b):
    return np.all(np.abs(b - a) < eps)


def assert_equal(a, b):
    if not nearly_equal(a, b):
        msg = 'this:\n'
        msg += repr(a)
        msg += '\nand that:\n'
        msg += repr(b)
        msg += '\nwere supposed to be equal but are not.'
        raise AssertionError(msg)


def test_cell_conv():
    # Make sure we get exactly zeros off-diagonal:
    assert (p2c([1, 1, 1, 90, 90, 90]) == np.eye(3)).all()

    # Constants
    a = 5.43
    d = a / 2.0
    h = a / np.sqrt(2.0)

    # Systems
    # Primitive cell, non-orthorhombic, non-cubic
    # Parameters
    si_prim_p = np.array([h] * 3 + [60.] * 3)
    # Tensor format
    si_prim_m = np.array([[0., d, d],
                          [d, 0., d],
                          [d, d, 0.]])
    # Tensor format in the default basis
    si_prim_m2 = np.array([[2.0, 0., 0.],
                           [1.0, np.sqrt(3.0), 0.],
                           [1.0, np.sqrt(3.0) / 3.0, 2 * np.sqrt(2 / 3)]])
    si_prim_m2 *= h / 2.0

    # Orthorhombic cell, non-cubic
    # Parameters
    si_ortho_p = np.array([h] * 2 + [a] + [90.] * 3)
    # Tensor format in the default basis
    si_ortho_m = np.array([[h, 0.0, 0.0],
                           [0.0, h, 0.0],
                           [0.0, 0.0, a]])

    # Cubic cell
    # Parameters
    si_cubic_p = np.array([a] * 3 + [90.] * 3)
    # Tensor format in the default basis
    si_cubic_m = np.array([[a, 0.0, 0.0],
                           [0.0, a, 0.0],
                           [0.0, 0.0, a]])

    # Cell matrix -> cell parameters
    assert_equal(c2p(si_prim_m), si_prim_p)
    assert_equal(c2p(si_prim_m2), si_prim_p)
    assert_equal(c2p(si_ortho_m), si_ortho_p)
    assert_equal(c2p(si_cubic_m), si_cubic_p)
    assert not nearly_equal(c2p(si_prim_m), si_ortho_p)

    # Cell parameters -> cell matrix
    assert_equal(p2c(si_prim_p), si_prim_m2)
    assert_equal(p2c(si_ortho_p), si_ortho_m)
    assert_equal(p2c(si_cubic_p), si_cubic_m)
    assert not nearly_equal(p2c(si_prim_p), si_ortho_m)

    # Idempotency (provided everything is provided in the default basis)
    ref1 = si_prim_m2[:]
    ref2 = si_ortho_m[:]
    ref3 = si_cubic_m[:]
    for _ in range(20):
        ref1[:] = p2c(c2p(ref1))
        ref2[:] = p2c(c2p(ref2))
        ref3[:] = p2c(c2p(ref3))

    assert_equal(ref1, si_prim_m2)
    assert_equal(ref2, si_ortho_m)
    assert_equal(ref3, si_cubic_m)



"""Functions to read from control file and from turbomole standard output"""

import os
import re
import subprocess
import warnings

import numpy as np

from ase import Atom, Atoms
from ase.calculators.calculator import ReadError
from ase.units import Bohr, Ha


def execute_command(args):
    """execute commands like sdg, eiger"""
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, encoding='ASCII')
    stdout, _stderr = proc.communicate()
    return stdout


def read_data_group(data_group):
    """read a turbomole data group from control file"""
    return execute_command(['sdg', data_group]).strip()


def parse_data_group(dgr, dg_name):
    """parse a data group"""
    if len(dgr) == 0:
        return None
    dg_key = '$' + dg_name
    if not dgr.startswith(dg_key):
        raise ValueError(f'data group does not start with {dg_key}')
    ndg = dgr.replace(dg_key, '').strip()
    ndg = re.sub(r'=\s+', '=', re.sub(r'\s+=', '=', ndg))
    if all(c not in ndg for c in ('\n', ' ', '=')):
        return ndg
    lsep = '\n' if '\n' in dgr else ' '
    result = {}
    lines = ndg.split(lsep)
    for line in lines:
        if len(line) == 0:
            continue
        ksep = '=' if '=' in line else None
        fields = line.strip().split(ksep)
        if len(fields) == 2:
            result[fields[0]] = fields[1]
        elif len(fields) == 1:
            result[fields[0]] = True
    return result


def read_output(regex, path):
    """collects all matching strings from the output"""
    hitlist = []
    checkfiles = []
    for filename in os.listdir(path):
        if filename.startswith('job.') or filename.endswith('.out'):
            checkfiles.append(filename)
    for filename in checkfiles:
        with open(filename) as f:
            lines = f.readlines()
            for line in lines:
                match = re.search(regex, line)
                if match:
                    hitlist.append(match.group(1))
    return hitlist


def read_version(path):
    """read the version from the tm output if stored in a file"""
    versions = read_output(r'TURBOMOLE\s+V(\d+\.\d+)\s+', path)
    if len(set(versions)) > 1:
        warnings.warn('different turbomole versions detected')
        version = list(set(versions))
    elif len(versions) == 0:
        warnings.warn('no turbomole version detected')
        version = None
    else:
        version = versions[0]
    return version


def read_datetime(path):
    """read the datetime of the most recent calculation
    from the tm output if stored in a file
    """
    datetimes = read_output(
        r'(\d{4}-[01]\d-[0-3]\d([T\s][0-2]\d:[0-5]'
        r'\d:[0-5]\d\.\d+)?([+-][0-2]\d:[0-5]\d|Z)?)', path)
    if len(datetimes) == 0:
        warnings.warn('no turbomole datetime detected')
        datetime = None
    else:
        # take the most recent time stamp
        datetime = sorted(datetimes, reverse=True)[0]
    return datetime


def read_runtime(path):
    """read the total runtime of calculations"""
    hits = read_output(r'total wall-time\s+:\s+(\d+.\d+)\s+seconds', path)
    if len(hits) == 0:
        warnings.warn('no turbomole runtimes detected')
        runtime = None
    else:
        runtime = np.sum([float(a) for a in hits])
    return runtime


def read_hostname(path):
    """read the hostname of the computer on which the calc has run"""
    hostnames = read_output(r'hostname is\s+(.+)', path)
    if len(set(hostnames)) > 1:
        warnings.warn('runs on different hosts detected')
        hostname = list(set(hostnames))
    else:
        hostname = hostnames[0]
    return hostname


def read_convergence(restart, parameters):
    """perform convergence checks"""
    if restart:
        if bool(len(read_data_group('restart'))):
            return False
        if bool(len(read_data_group('actual'))):
            return False
        if not bool(len(read_data_group('energy'))):
            return False
        if (os.path.exists('job.start') and
                os.path.exists('GEO_OPT_FAILED')):
            return False
        return True

    if parameters['task'] in ['optimize', 'geometry optimization']:
        if os.path.exists('GEO_OPT_CONVERGED'):
            return True
        elif os.path.exists('GEO_OPT_FAILED'):
            # check whether a failed scf convergence is the reason
            checkfiles = []
            for filename in os.listdir('.'):
                if filename.startswith('job.'):
                    checkfiles.append(filename)
            for filename in checkfiles:
                for line in open(filename):
                    if 'SCF FAILED TO CONVERGE' in line:
                        # scf did not converge in some jobex iteration
                        if filename == 'job.last':
                            raise RuntimeError('scf failed to converge')
                        else:
                            warnings.warn('scf failed to converge')
            warnings.warn('geometry optimization failed to converge')
            return False
        else:
            raise RuntimeError('error during geometry optimization')
    else:
        if os.path.isfile('dscf_problem'):
            raise RuntimeError('scf failed to converge')
        else:
            return True


def read_run_parameters(results):
    """read parameters set by define and not in self.parameters"""

    if 'calculation parameters' not in results.keys():
        results['calculation parameters'] = {}
    parameters = results['calculation parameters']
    dg = read_data_group('symmetry')
    parameters['point group'] = str(dg.split()[1])
    parameters['uhf'] = '$uhf' in read_data_group('uhf')
    # Gaussian function type
    gt = read_data_group('pople')
    if gt == '':
        parameters['gaussian type'] = 'spherical harmonic'
    else:
        gt = gt.split()[1]
        if gt == 'AO':
            parameters['gaussian type'] = 'spherical harmonic'
        elif gt == 'CAO':
            parameters['gaussian type'] = 'cartesian'
        else:
            parameters['gaussian type'] = None

    nvibro = read_data_group('nvibro')
    if nvibro:
        parameters['nuclear degrees of freedom'] = int(nvibro.split()[1])


def read_energy(results, post_HF):
    """Read energy from Turbomole energy file."""
    try:
        with open('energy') as enf:
            text = enf.read().lower()
    except OSError:
        raise ReadError('failed to read energy file')
    if text == '':
        raise ReadError('empty energy file')

    lines = iter(text.split('\n'))

    for line in lines:
        if line.startswith('$end'):
            break
        elif line.startswith('$'):
            pass
        else:
            energy_tmp = float(line.split()[1])
            if post_HF:
                energy_tmp += float(line.split()[4])
    # update energy units
    e_total = energy_tmp * Ha
    results['total energy'] = e_total


def read_occupation_numbers(results):
    """read occupation numbers with module 'eiger' """
    if 'molecular orbitals' not in results.keys():
        return
    mos = results['molecular orbitals']
    lines = execute_command(['eiger', '--all', '--pview']).split('\n')
    for line in lines:
        regex = (
            r'^\s+(\d+)\.\s([\sab])\s*(\d+)\s?(\w+)'
            r'\s+(\d*\.*\d*)\s+([-+]?\d+\.\d*)'
        )
        match = re.search(regex, line)
        if match:
            orb_index = int(match.group(3))
            if match.group(2) == 'a':
                spin = 'alpha'
            elif match.group(2) == 'b':
                spin = 'beta'
            else:
                spin = None
            ar_index = next(
                index for (index, molecular_orbital) in enumerate(mos)
                if (molecular_orbital['index'] == orb_index and
                    molecular_orbital['spin'] == spin)
            )
            mos[ar_index]['index by energy'] = int(match.group(1))
            irrep = str(match.group(4))
            mos[ar_index]['irreducible representation'] = irrep
            if match.group(5) != '':
                mos[ar_index]['occupancy'] = float(match.group(5))
            else:
                mos[ar_index]['occupancy'] = float(0)


def read_mos(results):
    """read the molecular orbital coefficients and orbital energies
    from files mos, alpha and beta"""

    results['molecular orbitals'] = []
    mos = results['molecular orbitals']
    keywords = ['scfmo', 'uhfmo_alpha', 'uhfmo_beta']
    spin = [None, 'alpha', 'beta']
    converged = None

    for index, keyword in enumerate(keywords):
        flen = None
        mo = {}
        orbitals_coefficients_line = []
        mo_string = read_data_group(keyword)
        if mo_string == '':
            continue
        mo_string += '\n$end'
        lines = mo_string.split('\n')
        for line in lines:
            if re.match(r'^\s*#', line):
                continue
            if 'eigenvalue' in line:
                if len(orbitals_coefficients_line) != 0:
                    mo['eigenvector'] = orbitals_coefficients_line
                    mos.append(mo)
                    mo = {}
                    orbitals_coefficients_line = []
                regex = (r'^\s*(\d+)\s+(\S+)\s+'
                         r'eigenvalue=([\+\-\d\.\w]+)\s')
                match = re.search(regex, line)
                mo['index'] = int(match.group(1))
                mo['irreducible representation'] = str(match.group(2))
                eig = float(re.sub('[dD]', 'E', match.group(3))) * Ha
                mo['eigenvalue'] = eig
                mo['spin'] = spin[index]
                mo['degeneracy'] = 1
                continue
            if keyword in line:
                # e.g. format(4d20.14)
                regex = r'format\(\d+[a-zA-Z](\d+)\.\d+\)'
                match = re.search(regex, line)
                if match:
                    flen = int(match.group(1))
                if ('scfdump' in line or 'expanded' in line or
                        'scfconv' not in line):
                    converged = False
                continue
            if '$end' in line:
                if len(orbitals_coefficients_line) != 0:
                    mo['eigenvector'] = orbitals_coefficients_line
                    mos.append(mo)
                break
            sfields = [line[i:i + flen]
                       for i in range(0, len(line), flen)]
            ffields = [float(f.replace('D', 'E').replace('d', 'E'))
                       for f in sfields]
            orbitals_coefficients_line += ffields
    return converged


def read_basis_set(results):
    """read the basis set"""
    results['basis set'] = []
    results['basis set formatted'] = {}
    bsf = read_data_group('basis')
    results['basis set formatted']['turbomole'] = bsf
    lines = bsf.split('\n')
    basis_set = {}
    functions = []
    function = {}
    primitives = []
    read_tag = False
    read_data = False
    for line in lines:
        if len(line.strip()) == 0:
            continue
        if '$basis' in line:
            continue
        if '$end' in line:
            break
        if re.match(r'^\s*#', line):
            continue
        if re.match(r'^\s*\*', line):
            if read_tag:
                read_tag = False
                read_data = True
            else:
                if read_data:
                    # end primitives
                    function['primitive functions'] = primitives
                    function['number of primitives'] = len(primitives)
                    primitives = []
                    functions.append(function)
                    function = {}
                    # end contracted
                    basis_set['functions'] = functions
                    functions = []
                    results['basis set'].append(basis_set)
                    basis_set = {}
                    read_data = False
                read_tag = True
            continue
        if read_tag:
            match = re.search(r'^\s*(\w+)\s+(.+)', line)
            if match:
                basis_set['element'] = match.group(1)
                basis_set['nickname'] = match.group(2)
            else:
                raise RuntimeError('error reading basis set')
        else:
            match = re.search(r'^\s+(\d+)\s+(\w+)', line)
            if match:
                if len(primitives) > 0:
                    # end primitives
                    function['primitive functions'] = primitives
                    function['number of primitives'] = len(primitives)
                    primitives = []
                    functions.append(function)
                    function = {}
                    # begin contracted
                function['shell type'] = str(match.group(2))
                continue
            regex = (
                r'^\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
                r'\s+([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
            )
            match = re.search(regex, line)
            if match:
                exponent = float(match.group(1))
                coefficient = float(match.group(3))
                primitives.append(
                    {'exponent': exponent, 'coefficient': coefficient}
                )


def read_ecps(results):
    """read the effective core potentials"""
    ecpf = read_data_group('ecp')
    if not bool(len(ecpf)):
        results['ecps'] = None
        results['ecps formatted'] = None
        return
    results['ecps'] = []
    results['ecps formatted'] = {}
    results['ecps formatted']['turbomole'] = ecpf
    lines = ecpf.split('\n')
    ecp = {}
    groups = []
    group = {}
    terms = []
    read_tag = False
    read_data = False
    for line in lines:
        if len(line.strip()) == 0:
            continue
        if '$ecp' in line:
            continue
        if '$end' in line:
            break
        if re.match(r'^\s*#', line):
            continue
        if re.match(r'^\s*\*', line):
            if read_tag:
                read_tag = False
                read_data = True
            else:
                if read_data:
                    # end terms
                    group['terms'] = terms
                    group['number of terms'] = len(terms)
                    terms = []
                    groups.append(group)
                    group = {}
                    # end group
                    ecp['groups'] = groups
                    groups = []
                    results['ecps'].append(ecp)
                    ecp = {}
                    read_data = False
                read_tag = True
            continue
        if read_tag:
            match = re.search(r'^\s*(\w+)\s+(.+)', line)
            if match:
                ecp['element'] = match.group(1)
                ecp['nickname'] = match.group(2)
            else:
                raise RuntimeError('error reading ecp')
        else:
            regex = r'ncore\s*=\s*(\d+)\s+lmax\s*=\s*(\d+)'
            match = re.search(regex, line)
            if match:
                ecp['number of core electrons'] = int(match.group(1))
                ecp['maximum angular momentum number'] = \
                    int(match.group(2))
                continue
            match = re.search(r'^(\w(\-\w)?)', line)
            if match:
                if len(terms) > 0:
                    # end terms
                    group['terms'] = terms
                    group['number of terms'] = len(terms)
                    terms = []
                    groups.append(group)
                    group = {}
                    # begin group
                group['title'] = str(match.group(1))
                continue
            regex = (r'^\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+'
                     r'(\d)\s+([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)')
            match = re.search(regex, line)
            if match:
                terms.append(
                    {
                        'coefficient': float(match.group(1)),
                        'power of r': float(match.group(3)),
                        'exponent': float(match.group(4))
                    }
                )


def read_forces(results, natoms):
    """Read forces from Turbomole gradient file."""
    dg = read_data_group('grad')
    if len(dg) == 0:
        return None
    with open('gradient') as file:
        lines = file.readlines()
    forces = np.array([[0, 0, 0]])

    nline = len(lines)
    iline = -1

    for i in range(nline):
        if 'cycle' in lines[i]:
            iline = i

    if iline < 0:
        raise RuntimeError('Please check TURBOMOLE gradients')

    # next line
    iline += natoms + 1
    # $end line
    nline -= 1
    # read gradients
    for i in range(iline, nline):
        line = lines[i].replace('D', 'E')
        tmp = np.array([[float(f) for f in line.split()[0:3]]])
        forces = np.concatenate((forces, tmp))
    # Note the '-' sign for turbomole, to get forces
    forces = -np.delete(forces, np.s_[0:1], axis=0) * Ha / Bohr
    results['energy gradient'] = (-forces).tolist()
    return forces


def read_gradient(results):
    """read all information in file 'gradient'"""
    grad_string = read_data_group('grad')
    if len(grad_string) == 0:
        return
#       try to reuse ase:
#       structures = read('gradient', index=':')
    lines = grad_string.split('\n')
    history = []
    image = {}
    gradient = []
    atoms = Atoms()
    (cycle, energy, norm) = (None, None, None)
    for line in lines:
        # cycle lines
        regex = (
            r'^\s*cycle =\s*(\d+)\s+'
            r'SCF energy =\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+'
            r'\|dE\/dxyz\| =\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
        )
        match = re.search(regex, line)
        if match:
            if len(atoms):
                image['optimization cycle'] = cycle
                image['total energy'] = energy
                image['gradient norm'] = norm
                image['energy gradient'] = gradient
                history.append(image)
                image = {}
                atoms = Atoms()
                gradient = []
            cycle = int(match.group(1))
            energy = float(match.group(2)) * Ha
            norm = float(match.group(4)) * Ha / Bohr
            continue
        # coordinate lines
        regex = (
            r'^\s*([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            r'\s+(\w+)'
        )
        match = re.search(regex, line)
        if match:
            x = float(match.group(1)) * Bohr
            y = float(match.group(3)) * Bohr
            z = float(match.group(5)) * Bohr
            symbol = str(match.group(7)).capitalize()

            if symbol == 'Q':
                symbol = 'X'
            atoms += Atom(symbol, (x, y, z))

            continue
        # gradient lines
        regex = (
            r'^\s*([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
        )
        match = re.search(regex, line)
        if match:
            gradx = float(match.group(1).replace('D', 'E')) * Ha / Bohr
            grady = float(match.group(3).replace('D', 'E')) * Ha / Bohr
            gradz = float(match.group(5).replace('D', 'E')) * Ha / Bohr
            gradient.append([gradx, grady, gradz])

    image['optimization cycle'] = cycle
    image['total energy'] = energy
    image['gradient norm'] = norm
    image['energy gradient'] = gradient
    history.append(image)
    results['geometry optimization history'] = history


def read_hessian(results, noproj=False):
    """Read in the hessian matrix"""
    results['hessian matrix'] = {}
    results['hessian matrix']['array'] = []
    results['hessian matrix']['units'] = '?'
    results['hessian matrix']['projected'] = True
    results['hessian matrix']['mass weighted'] = True
    dg = read_data_group('nvibro')
    if len(dg) == 0:
        return
    nvibro = int(dg.split()[1])
    results['hessian matrix']['dimension'] = nvibro
    row = []
    key = 'hessian'
    if noproj:
        key = 'npr' + key
        results['hessian matrix']['projected'] = False
    lines = read_data_group(key).split('\n')
    for line in lines:
        if key in line:
            continue
        fields = line.split()
        row.extend(fields[2:len(fields)])
        if len(row) == nvibro:
            # check whether it is mass-weighted
            float_row = [float(element) for element in row]
            results['hessian matrix']['array'].append(float_row)
            row = []


def read_normal_modes(results, noproj=False):
    """Read in vibrational normal modes"""
    results['normal modes'] = {}
    results['normal modes']['array'] = []
    results['normal modes']['projected'] = True
    results['normal modes']['mass weighted'] = True
    results['normal modes']['units'] = '?'
    dg = read_data_group('nvibro')
    if len(dg) == 0:
        return
    nvibro = int(dg.split()[1])
    results['normal modes']['dimension'] = nvibro
    row = []
    key = 'vibrational normal modes'
    if noproj:
        key = 'npr' + key
        results['normal modes']['projected'] = False
    lines = read_data_group(key).split('\n')
    for line in lines:
        if key in line:
            continue
        if '$end' in line:
            break
        fields = line.split()
        row.extend(fields[2:len(fields)])
        if len(row) == nvibro:
            # check whether it is mass-weighted
            float_row = [float(element) for element in row]
            results['normal modes']['array'].append(float_row)
            row = []


def read_vibrational_reduced_masses(results):
    """Read vibrational reduced masses"""
    results['vibrational reduced masses'] = []
    dg = read_data_group('vibrational reduced masses')
    if len(dg) == 0:
        return
    lines = dg.split('\n')
    for line in lines:
        if '$vibrational' in line:
            continue
        if '$end' in line:
            break
        fields = [float(element) for element in line.split()]
        results['vibrational reduced masses'].extend(fields)


def read_vibrational_spectrum(results, noproj=False):
    """Read the vibrational spectrum"""
    results['vibrational spectrum'] = []
    key = 'vibrational spectrum'
    if noproj:
        key = 'npr' + key
    lines = read_data_group(key).split('\n')
    for line in lines:
        dictionary = {}
        regex = (
            r'^\s+(\d+)\s+(\S*)\s+([-+]?\d+\.\d*)'
            r'\s+(\d+\.\d*)\s+(\S+)\s+(\S+)'
        )
        match = re.search(regex, line)
        if match:
            dictionary['mode number'] = int(match.group(1))
            dictionary['irreducible representation'] = str(match.group(2))
            dictionary['frequency'] = {
                'units': 'cm^-1',
                'value': float(match.group(3))
            }
            dictionary['infrared intensity'] = {
                'units': 'km/mol',
                'value': float(match.group(4))
            }

            if match.group(5) == 'YES':
                dictionary['infrared active'] = True
            elif match.group(5) == 'NO':
                dictionary['infrared active'] = False
            else:
                dictionary['infrared active'] = None

            if match.group(6) == 'YES':
                dictionary['Raman active'] = True
            elif match.group(6) == 'NO':
                dictionary['Raman active'] = False
            else:
                dictionary['Raman active'] = None

            results['vibrational spectrum'].append(dictionary)


def read_ssquare(results):
    """Read the expectation value of S^2 operator"""
    s2_string = read_data_group('ssquare from dscf')
    if s2_string == '':
        return
    string = s2_string.split('\n')[1]
    ssquare = float(re.search(r'^\s*(\d+\.*\d*)', string).group(1))
    results['ssquare from scf calculation'] = ssquare


def read_dipole_moment(results):
    """Read the dipole moment"""
    dip_string = read_data_group('dipole')
    if dip_string == '':
        return
    lines = dip_string.split('\n')
    for line in lines:
        regex = (
            r'^\s+x\s+([-+]?\d+\.\d*)\s+y\s+([-+]?\d+\.\d*)'
            r'\s+z\s+([-+]?\d+\.\d*)\s+a\.u\.'
        )
        match = re.search(regex, line)
        if match:
            dip_vec = [float(match.group(c)) for c in range(1, 4)]
        regex = r'^\s+\| dipole \| =\s+(\d+\.*\d*)\s+debye'
        match = re.search(regex, line)
        if match:
            dip_abs_val = float(match.group(1))
    results['electric dipole moment'] = {}
    results['electric dipole moment']['vector'] = {
        'array': dip_vec,
        'units': 'a.u.'
    }
    results['electric dipole moment']['absolute value'] = {
        'value': dip_abs_val,
        'units': 'Debye'
    }


def read_charges(filename, natoms):
    """read partial charges on atoms from an ESP fit"""
    charges = None
    if os.path.exists(filename):
        with open(filename) as infile:
            lines = infile.readlines()
        oklines = None
        for n, line in enumerate(lines):
            if 'atom  radius/au   charge' in line:
                oklines = lines[n + 1:n + natoms + 1]
        if oklines is not None:
            qm_charges = [float(line.split()[3]) for line in oklines]
            charges = np.array(qm_charges)
    return charges


def read_point_charges():
    """read point charges from previous calculation"""
    pcs = read_data_group('point_charges')
    lines = pcs.split('\n')[1:]
    (charges, positions) = ([], [])
    for line in lines:
        columns = [float(col) for col in line.strip().split()]
        positions.append([col * Bohr for col in columns[0:3]])
        charges.append(columns[3])
    return charges, positions



from ase.calculators.aims import get_aims_version

version_string = """\
          Invoking FHI-aims ...

blah blah blah

  FHI-aims version      : 200112.2
  Commit number         : GITDIR-NOTFOUND
"""


def test_get_aims_version():
    assert get_aims_version(version_string) == '200112.2'



import numpy as np
import pytest
from pytest import mark

from ase import Atoms


@mark.calculator_lite
def test_modify_parameters(KIM):
    """
    Check that the KIM calculator is capable of retrieving and updating model
    parameters correctly.  This is done by instantiating the calculator for a
    specific Lennard-Jones (LJ) potential, included with the KIM API, with a
    known cutoff for Mo-Mo interactions.  An Mo dimer is then constructed with
    a random separation that falls within the cutoff and its energy using the
    original potential parameters is computed.  Next, the original value of the
    "epsilon" parameter for Mo is retrieved.  The value of the parameter is
    then set to a scaling factor times the original value and the energy
    recomputed.  In the Lennard-Jones potential, the energy is directly
    proportional to the value of parameter "epsilon"; thus, the final energy
    computed is asserted to be approximately equal to the scaling factor times
    the original energy.
    """

    # In LennardJones612_UniversalShifted__MO_959249795837_003, the cutoff
    # for Mo interaction is 10.9759 Angstroms.
    cutoff = 10.9759

    # Create random dimer with separation < cutoff
    dimer_separation = np.random.RandomState(
        11).uniform(0.1 * cutoff, 0.6 * cutoff)
    atoms = Atoms("Mo" * 2, positions=[[0, 0, 0], [0, 0, dimer_separation]])

    calc = KIM("LennardJones612_UniversalShifted__MO_959249795837_003")
    atoms.calc = calc

    # Retrieve the original energy scaling parameter
    eps_orig = calc.get_parameters(epsilons=4879)["epsilons"][1]  # eV

    # Get the energy using the original parameter as a reference value
    E_orig = atoms.get_potential_energy()  # eV

    # Scale the energy scaling parameter and set this value to the calculator
    energy_scaling_factor = 2.0
    eps_modified = energy_scaling_factor * eps_orig
    calc.set_parameters(epsilons=[4879, eps_modified])

    # Get the energy after modifying the parameter
    E_modified = atoms.get_potential_energy()  # eV

    assert E_modified == pytest.approx(energy_scaling_factor * E_orig, rel=1e-4)



import pytest
from numpy.testing import assert_allclose

from ase.calculators.fd import calculate_numerical_forces
from ase.cluster.icosahedron import Icosahedron
from ase.data import atomic_masses, atomic_numbers
from ase.optimize import LBFGS


@pytest.fixture()
def ar_nc():
    ar_nc = Icosahedron('Ar', noshells=2)
    ar_nc.cell = [[300, 0, 0], [0, 300, 0], [0, 0, 300]]
    ar_nc.pbc = True
    return ar_nc


@pytest.fixture()
def params():
    params = {}
    params['pair_style'] = 'lj/cut 8.0'
    params['pair_coeff'] = ['1 1 0.0108102 3.345']
    params['masses'] = ['1 {}'.format(atomic_masses[atomic_numbers['Ar']])]
    return params


@pytest.mark.calculator_lite()
@pytest.mark.calculator('lammpsrun')
def test_Ar_minimize(factory, ar_nc, params):
    with factory.calc(specorder=['Ar'], **params) as calc:
        ar_nc.calc = calc

        assert_allclose(ar_nc.get_potential_energy(), -0.468147667942117,
                        atol=1e-4, rtol=1e-4)
        assert_allclose(ar_nc.get_forces(),
                        calculate_numerical_forces(ar_nc),
                        atol=1e-4, rtol=1e-4)

        with LBFGS(ar_nc) as dyn:
            dyn.run(fmax=1E-6)

        assert_allclose(ar_nc.get_potential_energy(), -0.4791815886953914,
                        atol=1e-4, rtol=1e-4)
        assert_allclose(ar_nc.get_forces(),
                        calculate_numerical_forces(ar_nc),
                        atol=1e-4, rtol=1e-4)


@pytest.mark.calculator_lite()
@pytest.mark.calculator('lammpsrun')
def test_Ar_minimize_multistep(factory, ar_nc, params):
    ar_nc = Icosahedron('Ar', noshells=2)
    ar_nc.cell = [[300, 0, 0], [0, 300, 0], [0, 0, 300]]
    ar_nc.pbc = True

    with factory.calc(specorder=['Ar'], **params) as calc:
        ar_nc.calc = calc
        F1_numer = calculate_numerical_forces(ar_nc)

        assert_allclose(ar_nc.get_potential_energy(), -0.468147667942117,
                        atol=1e-4, rtol=1e-4)
        assert_allclose(ar_nc.get_forces(), F1_numer,
                        atol=1e-4, rtol=1e-4)

        calc.set(minimize='1.0e-15 1.0e-6 2000 4000')

        # set_atoms=True to read final coordinates after minimization
        calc.run(set_atoms=True)

        # get final coordinates after minimization
        ar_nc.set_positions(calc.atoms.positions)

        assert_allclose(ar_nc.get_potential_energy(), -0.4791815887032201,
                        atol=1e-4, rtol=1e-4)
        assert_allclose(ar_nc.get_forces(),
                        calculate_numerical_forces(ar_nc),
                        atol=1e-4, rtol=1e-4)



import numpy as np
from pytest import mark

from ase import Atoms
from ase.calculators.fd import (
    calculate_numerical_forces,
    calculate_numerical_stress,
)


@mark.calculator_lite
def test_multi_neighlist(KIM, testdir):
    """
    To test that the correct energy/forces/stress can be computed using a
    model that implements multiple cutoffs.  This is done by construct a 10
    Angstrom x 10 Angstrom x 10 Angstrom non-periodic cell filled with 15
    randomly positioned atoms and requesting tha tthe model compute the
    energy, forces, and virial stress.  The energy returned by the model is
    compared to a known precomputed value, while the forces and stress
    returned are compared to numerical estimates via finite difference.
    """
    # Create random cluster of atoms
    positions = np.random.RandomState(34).rand(15, 3) * 10
    atoms = Atoms(
        "Ar" * 15, positions=positions, pbc=False,
        cell=[[10, 0, 0], [0, 10, 0], [0, 0, 10]]
    )

    calc = KIM("ex_model_Ar_P_Morse_MultiCutoff")
    atoms.calc = calc

    # Get energy and analytical forces/stress from KIM Model
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    stress = atoms.get_stress()

    # Previously computed energy for this configuration for this model
    energy_ref = 34.69963483186903  # eV

    # Compute forces and virial stress numerically
    forces_numer = calculate_numerical_forces(atoms, eps=0.0001)
    stress_numer = calculate_numerical_stress(atoms, eps=0.0001, voigt=True)

    tol = 1e-6
    assert np.isclose(energy, energy_ref, tol)
    assert np.allclose(forces, forces_numer, tol)
    assert np.allclose(stress, stress_numer, tol)



import sys
from subprocess import run

import pytest

from ase.parallel import world


def test_mpi():
    x = 42.0
    with pytest.warns(FutureWarning):
        x = world.sum(x)
    x = world.sum_scalar(x)
    assert x == 42 * world.size


@pytest.mark.skip(reason='Does not work and no time to investigate.')
def test_mpi_unused_on_import():
    """Try to import all ASE modules and check that ase.parallel.world has not
    been used.  We want to delay use of world until after MPI4PY has been
    imported.

    We run the test in a subprocess so that we have a clean Python
    interpreter."""

    # Should cover most of ASE:
    modules = ['ase.optimize',
               'ase.db',
               'ase.gui']

    imports = 'import ' + ', '.join(modules)

    run([sys.executable,
         '-c',
         '{imports}; from ase.parallel import world; assert world.comm is None'
         .format(imports=imports)],
        check=True)



import numpy as np

from ase import Atoms
from ase.build import bulk, make_supercell


def test_bulk():
    a1 = bulk('ZnS', 'wurtzite', a=3.0, u=0.23) * (1, 2, 1)
    a2 = bulk('ZnS', 'wurtzite', a=3.0, u=0.23, orthorhombic=True)
    a1.cell = a2.cell
    a1.wrap()
    assert abs(a1.positions - a2.positions).max() < 1e-14


class TestCubic:
    def test_sc(self):
        name = "Po"
        structure = "sc"
        a = 1.0
        atoms0 = bulk(name, structure, a=a)
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_bcc(self):
        name = "Li"
        structure = "bcc"
        a = 1.0
        P = [[0, 1, 1], [1, 0, 1], [1, 1, 0]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_cesiumchloride(self):
        name = "CsCl"
        structure = "cesiumchloride"
        a = 1.0
        atoms0 = bulk(name, structure, a=a)
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_fcc(self):
        name = "Cu"
        structure = "fcc"
        a = 1.0
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms0 = atoms0[[0, 3, 2, 1]]  # sort
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_diamond(self):
        name = "C"
        structure = "diamond"
        a = 1.0
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms0 = atoms0[[0, 1, 6, 7, 4, 5, 2, 3]]  # sort
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_zincblende(self):
        name = "ZnS"
        structure = "zincblende"
        a = 1.0
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms0 = atoms0[[0, 1, 6, 7, 4, 5, 2, 3]]  # sort
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_rocksalt(self):
        name = "NaCl"
        structure = "rocksalt"
        a = 1.0
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms0 = Atoms(atoms0[[0, 1, 6, 7, 4, 5, 2, 3]])  # sort
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def test_fluorite(self):
        name = "CaF2"
        structure = "fluorite"
        a = 1.0
        P = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
        atoms0 = make_supercell(bulk(name, structure, a=a), P)
        atoms0 = atoms0[[0, 1, 2, 9, 10, 11, 6, 7, 8, 3, 4, 5]]  # sort
        atoms1 = bulk(name, structure, a=a, cubic=True)
        self.compare(atoms0, atoms1)

    def compare(self, atoms0: Atoms, atoms1: Atoms):
        a = 1e-14
        print(atoms0.positions)
        print(atoms1.positions)
        np.testing.assert_allclose(atoms0.cell, atoms1.cell, atol=a)
        np.testing.assert_allclose(atoms0.positions, atoms1.positions, atol=a)


def hasmom(*args, **kwargs):
    return bulk(*args, **kwargs).has('initial_magmoms')


def test_magnetic_or_not():
    assert hasmom('Fe')
    assert hasmom('Fe', orthorhombic=True)
    assert hasmom('Fe', cubic=True)
    assert hasmom('Fe', 'bcc', 4.0)
    assert not hasmom('Fe', 'fcc', 4.0)
    assert not hasmom('Ti')
    assert not hasmom('Ti', 'bcc', 4.0)

    assert hasmom('Co')
    assert hasmom('Ni')



from io import StringIO

from ase.io import read, write


def test_jsondb():
    """Read and write json from/to file descriptor."""

    s = """
    {"1":
         {"numbers": [1, 1],
          "positions": [[0.0, 0.0, 0.35],
                        [0.0, 0.0, -0.35]]}}
    """

    fd = StringIO(s)
    a = read(fd, format='json')
    assert a.get_chemical_formula() == 'H2'

    fd = StringIO()
    write(fd, a, format='json')

    fd.seek(0)
    a = read(fd, format='json')
    assert a.get_chemical_formula() == 'H2'



# Note:
# Try to avoid module level import statements here to reduce
# import time during CLI execution


class CLICommand:
    """Analyze NEB trajectories by making band plots.

    One file:

        ase nebplot neb.traj

    Multiple files:

        ase nebplot neb1.traj neb2.traj

    Specify output:

        ase nebplot neb1.traj neb2.traj myfile.pdf
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('filenames', nargs='+',
            help='one or more trajectory files to analyze')
        add('output', nargs='?',
            help='optional name of output file, default=nebplots.pdf')
        add('--nimages', dest='n_images', type=int, default=None,
            help='number of images per band, guessed if not supplied')
        add('--share-x', dest='constant_x', action='store_true',
            help='use a single x axis scale for all plots')
        add('--share-y', dest='constant_y', action='store_true',
            help='use a single y axis scale for all plots')

    @staticmethod
    def run(args, parser):
        from ase.gui.images import Images
        from ase.mep import NEBTools

        # Nothing will ever be stored in args.output; need to manually find
        # if its supplied by checking extensions.
        if args.filenames[-1].endswith('.pdf'):
            args.output = args.filenames.pop(-1)
        else:
            args.output = 'nebplots.pdf'

        images = Images()
        images.read(args.filenames)
        nebtools = NEBTools(images=images)
        nebtools.plot_bands(constant_x=args.constant_x,
                            constant_y=args.constant_y,
                            nimages=args.n_images,
                            label=args.output[:-4])



"""
Module for povray file format support.

See http://www.povray.org/ for details on the format.
"""
from collections.abc import Sequence
from os import unlink
from pathlib import Path
from subprocess import DEVNULL, check_call

import numpy as np

from ase import Atoms
from ase.constraints import FixAtoms
from ase.io.utils import PlottingVariables


def pa(array):
    """Povray array syntax"""
    return '<' + ', '.join(f"{x:>6.2f}" for x in tuple(array)) + '>'


def pc(array):
    """Povray color syntax"""
    if isinstance(array, str):
        return 'color ' + array
    if isinstance(array, float):
        return f'rgb <{array:.2f}>*3'.format(array)
    L = len(array)
    if L > 2 and L < 6:
        return f"rgb{'' if L == 3 else 't' if L == 4 else 'ft'} <" +\
            ', '.join(f"{x:.2f}" for x in tuple(array)) + '>'


def get_bondpairs(atoms, radius=1.1):
    """Get all pairs of bonding atoms

    Return all pairs of atoms which are closer than radius times the
    sum of their respective covalent radii.  The pairs are returned as
    tuples::

      (a, b, (i1, i2, i3))

    so that atoms a bonds to atom b displaced by the vector::

        _     _     _
      i c + i c + i c ,
       1 1   2 2   3 3

    where c1, c2 and c3 are the unit cell vectors and i1, i2, i3 are
    integers."""

    from ase.data import covalent_radii
    from ase.neighborlist import NeighborList
    cutoffs = radius * covalent_radii[atoms.numbers]
    nl = NeighborList(cutoffs=cutoffs, self_interaction=False)
    nl.update(atoms)
    bondpairs = []
    for a in range(len(atoms)):
        indices, offsets = nl.get_neighbors(a)
        bondpairs.extend([(a, a2, offset)
                          for a2, offset in zip(indices, offsets)])
    return bondpairs


def set_high_bondorder_pairs(bondpairs, high_bondorder_pairs=None):
    """Set high bondorder pairs

    Modify bondpairs list (from get_bondpairs((atoms)) to include high
    bondorder pairs.

    Parameters:
    -----------
    bondpairs: List of pairs, generated from get_bondpairs(atoms)
    high_bondorder_pairs: Dictionary of pairs with high bond orders
                          using the following format:
                          { ( a1, b1 ): ( offset1, bond_order1, bond_offset1),
                            ( a2, b2 ): ( offset2, bond_order2, bond_offset2),
                            ...
                          }
                          offset, bond_order, bond_offset are optional.
                          However, if they are provided, the 1st value is
                          offset, 2nd value is bond_order,
                          3rd value is bond_offset """

    if high_bondorder_pairs is None:
        high_bondorder_pairs = {}
    bondpairs_ = []
    for pair in bondpairs:
        (a, b) = (pair[0], pair[1])
        if (a, b) in high_bondorder_pairs.keys():
            bondpair = [a, b] + [item for item in high_bondorder_pairs[(a, b)]]
            bondpairs_.append(bondpair)
        elif (b, a) in high_bondorder_pairs.keys():
            bondpair = [a, b] + [item for item in high_bondorder_pairs[(b, a)]]
            bondpairs_.append(bondpair)
        else:
            bondpairs_.append(pair)
    return bondpairs_


class POVRAY:
    # These new styles were an attempt to port the old styles o the correct
    # gamma, many had or still have unphysical light properties inorder to
    # acheive a certain look.
    material_styles_dict = dict(
        simple='finish {phong 0.7 ambient 0.4 diffuse 0.55}',
        # In general, 'pale' doesn't conserve energy and can look
        # strange in many cases.
        pale=('finish {ambient 0.9 diffuse 0.30 roughness 0.001 '
              'specular 0.2 }'),
        intermediate=('finish {ambient 0.4 diffuse 0.6 specular 0.1 '
                      'roughness 0.04}'),
        vmd=(
            'finish {ambient 0.2 diffuse 0.80 phong 0.25 phong_size 10.0 '
            'specular 0.2 roughness 0.1}'),
        jmol=('finish {ambient 0.4 diffuse 0.6 specular 1 roughness 0.001 '
              'metallic}'),
        ase2=('finish {ambient 0.2 brilliance 3 diffuse 0.6 metallic '
              'specular 0.7 roughness 0.04 reflection 0.15}'),
        ase3=('finish {ambient 0.4 brilliance 2 diffuse 0.6 metallic '
              'specular 1.0 roughness 0.001 reflection 0.0}'),
        glass=('finish {ambient 0.4 diffuse 0.35 specular 1.0 '
               'roughness 0.001}'),
        glass2=('finish {ambient 0.3 diffuse 0.3 specular 1.0 '
                'reflection 0.25 roughness 0.001}'),
    )

    # These styles were made when assumed_gamma was 1.0 which gives poor color
    # reproduction, the correct gamma is 2.2 for the sRGB standard.
    material_styles_dict_old = dict(
        simple='finish {phong 0.7}',
        pale=('finish {ambient 0.5 diffuse 0.85 roughness 0.001 '
              'specular 0.200 }'),
        intermediate=('finish {ambient 0.3 diffuse 0.6 specular 0.1 '
                      'roughness 0.04}'),
        vmd=('finish {ambient 0.0 diffuse 0.65 phong 0.1 phong_size 40.0 '
             'specular 0.5 }'),
        jmol=('finish {ambient 0.2 diffuse 0.6 specular 1 roughness 0.001 '
              'metallic}'),
        ase2=('finish {ambient 0.05 brilliance 3 diffuse 0.6 metallic '
              'specular 0.7 roughness 0.04 reflection 0.15}'),
        ase3=('finish {ambient 0.15 brilliance 2 diffuse 0.6 metallic '
              'specular 1.0 roughness 0.001 reflection 0.0}'),
        glass=('finish {ambient 0.05 diffuse 0.3 specular 1.0 '
               'roughness 0.001}'),
        glass2=('finish {ambient 0.01 diffuse 0.3 specular 1.0 '
                'reflection 0.25 roughness 0.001}'),
    )

    def __init__(self, cell, cell_vertices, positions, diameters, colors,
                 image_width, image_height, constraints=(), isosurfaces=[],
                 display=False, pause=True, transparent=True, canvas_width=None,
                 canvas_height=None, camera_dist=50., image_plane=None,
                 camera_type='orthographic', point_lights=[],
                 area_light=[(2., 3., 40.), 'White', .7, .7, 3, 3],
                 background='White', textures=None, transmittances=None,
                 depth_cueing=False, cue_density=5e-3,
                 celllinewidth=0.05, bondlinewidth=0.10, bondatoms=[],
                 exportconstraints=False):
        """
        # x, y is the image plane, z is *out* of the screen
        cell: ase.cell
            cell object
        cell_vertices: 2-d numpy array
            contains the 8 vertices of the cell, each with three coordinates
        positions: 2-d numpy array
            number of atoms length array with three coordinates for positions
        diameters: 1-d numpy array
            diameter of atoms (in order with positions)
        colors: list of str
            colors of atoms (in order with positions)
        image_width: float
            image width in pixels
        image_height: float
            image height in pixels
        constraints: Atoms.constraints
            constraints to be visualized
        isosurfaces: list of POVRAYIsosurface
            composite object to write/render POVRAY isosurfaces
        display: bool
            display while rendering
        pause: bool
            pause when done rendering (only if display)
        transparent: bool
            make background transparent
        canvas_width: int
            width of canvas in pixels
        canvas_height: int
            height of canvas in pixels
        camera_dist: float
            distance from camera to front atom
        image_plane: float
            distance from front atom to image plane
        camera_type: str
            if 'orthographic' perspective, ultra_wide_angle
        point_lights: list of 2-element sequences
            like [[loc1, color1], [loc2, color2],...]
        area_light: 3-element sequence of location (3-tuple), color (str),
                   width (float), height (float),
                   Nlamps_x (int), Nlamps_y (int)
            example [(2., 3., 40.), 'White', .7, .7, 3, 3]
        background: str
            color specification, e.g., 'White'
        textures: list of str
            length of atoms list of texture names
        transmittances: list of floats
            length of atoms list of transmittances of the atoms
        depth_cueing: bool
            whether or not to use depth cueing a.k.a. fog
            use with care - adjust the camera_distance to be closer
        cue_density: float
            if there is depth_cueing, how dense is it (how dense is the fog)
        celllinewidth: float
            radius of the cylinders representing the cell (Ang.)
        bondlinewidth: float
            radius of the cylinders representing bonds (Ang.)
        bondatoms: list of lists (polymorphic)
            [[atom1, atom2], ... ] pairs of bonding atoms
             For bond order > 1 = [[atom1, atom2, offset,
                                    bond_order, bond_offset],
                                   ... ]
             bond_order: 1, 2, 3 for single, double,
                          and triple bond
             bond_offset: vector for shifting bonds from
                           original position. Coordinates are
                           in Angstrom unit.
        exportconstraints: bool
            honour FixAtoms and mark?"""

        # attributes from initialization
        self.area_light = area_light
        self.background = background
        self.bondatoms = bondatoms
        self.bondlinewidth = bondlinewidth
        self.camera_dist = camera_dist
        self.camera_type = camera_type
        self.celllinewidth = celllinewidth
        self.cue_density = cue_density
        self.depth_cueing = depth_cueing
        self.display = display
        self.exportconstraints = exportconstraints
        self.isosurfaces = isosurfaces
        self.pause = pause
        self.point_lights = point_lights
        self.textures = textures
        self.transmittances = transmittances
        self.transparent = transparent

        self.image_width = image_width
        self.image_height = image_height
        self.colors = colors
        self.cell = cell
        self.diameters = diameters

        # calculations based on passed inputs

        z0 = positions[:, 2].max()
        self.offset = (image_width / 2, image_height / 2, z0)
        self.positions = positions - self.offset

        if cell_vertices is not None:
            self.cell_vertices = cell_vertices - self.offset
            self.cell_vertices.shape = (2, 2, 2, 3)
        else:
            self.cell_vertices = None

        ratio = float(self.image_width) / self.image_height
        if canvas_width is None:
            if canvas_height is None:
                self.canvas_width = min(self.image_width * 15, 640)
                self.canvas_height = min(self.image_height * 15, 640)
            else:
                self.canvas_width = canvas_height * ratio
                self.canvas_height = canvas_height
        elif canvas_height is None:
            self.canvas_width = canvas_width
            self.canvas_height = self.canvas_width / ratio
        else:
            raise RuntimeError("Can't set *both* width and height!")

        # Distance to image plane from camera
        if image_plane is None:
            if self.camera_type == 'orthographic':
                self.image_plane = 1 - self.camera_dist
            else:
                self.image_plane = 0
        self.image_plane += self.camera_dist

        self.constrainatoms = []
        for c in constraints:
            if isinstance(c, FixAtoms):
                # self.constrainatoms.extend(c.index) # is this list-like?
                for n, i in enumerate(c.index):
                    self.constrainatoms += [i]

    @classmethod
    def from_PlottingVariables(cls, pvars, **kwargs):
        cell = pvars.cell
        cell_vertices = pvars.cell_vertices
        if 'colors' in kwargs:
            colors = kwargs.pop('colors')
        else:
            colors = pvars.colors
        diameters = pvars.d
        image_height = pvars.h
        image_width = pvars.w
        positions = pvars.positions
        constraints = pvars.constraints
        return cls(cell=cell, cell_vertices=cell_vertices, colors=colors,
                   constraints=constraints, diameters=diameters,
                   image_height=image_height, image_width=image_width,
                   positions=positions, **kwargs)

    @classmethod
    def from_atoms(cls, atoms, **kwargs):
        return cls.from_plotting_variables(
            PlottingVariables(atoms, scale=1.0), **kwargs)

    def write_ini(self, path):
        """Write ini file."""

        ini_str = f"""\
Input_File_Name={path.with_suffix('.pov').name}
Output_to_File=True
Output_File_Type=N
Output_Alpha={'on' if self.transparent else 'off'}
; if you adjust Height, and width, you must preserve the ratio
; Width / Height = {self.canvas_width / self.canvas_height:f}
Width={self.canvas_width}
Height={self.canvas_height}
Antialias=True
Antialias_Threshold=0.1
Display={self.display}
Display_Gamma=2.2
Pause_When_Done={self.pause}
Verbose=False
"""
        with open(path, 'w') as fd:
            fd.write(ini_str)
        return path

    def write_pov(self, path):
        """Write pov file."""

        point_lights = '\n'.join(f"light_source {{{pa(loc)} {pc(rgb)}}}"
                                 for loc, rgb in self.point_lights)

        area_light = ''
        if self.area_light is not None:
            loc, color, width, height, nx, ny = self.area_light
            area_light += f"""\nlight_source {{{pa(loc)} {pc(color)}
  area_light <{width:.2f}, 0, 0>, <0, {height:.2f}, 0>, {nx:n}, {ny:n}
  adaptive 1 jitter}}"""

        fog = ''
        if self.depth_cueing and (self.cue_density >= 1e-4):
            # same way vmd does it
            if self.cue_density > 1e4:
                # larger does not make any sense
                dist = 1e-4
            else:
                dist = 1. / self.cue_density
            fog += f'fog {{fog_type 1 distance {dist:.4f} '\
                   f'color {pc(self.background)}}}'

        mat_style_keys = (f'#declare {k} = {v}'
                          for k, v in self.material_styles_dict.items())
        mat_style_keys = '\n'.join(mat_style_keys)

        # Draw unit cell
        cell_vertices = ''
        if self.cell_vertices is not None:
            for c in range(3):
                for j in ([0, 0], [1, 0], [1, 1], [0, 1]):
                    p1 = self.cell_vertices[tuple(j[:c]) + (0,) + tuple(j[c:])]
                    p2 = self.cell_vertices[tuple(j[:c]) + (1,) + tuple(j[c:])]

                    distance = np.linalg.norm(p2 - p1)
                    if distance < 1e-12:
                        continue

                    cell_vertices += f'cylinder {{{pa(p1)}, {pa(p2)}, '\
                                     'Rcell pigment {Black}}\n'
                    # all strings are f-strings for consistency
            cell_vertices = cell_vertices.strip('\n')

        # Draw atoms
        a = 0
        atoms = ''
        for loc, dia, col in zip(self.positions, self.diameters, self.colors):
            tex = 'ase3'
            trans = 0.
            if self.textures is not None:
                tex = self.textures[a]
            if self.transmittances is not None:
                trans = self.transmittances[a]
            atoms += f'atom({pa(loc)}, {dia / 2.:.2f}, {pc(col)}, '\
                     f'{trans}, {tex}) // #{a:n}\n'
            a += 1
        atoms = atoms.strip('\n')

        # Draw atom bonds
        bondatoms = ''
        for pair in self.bondatoms:
            # Make sure that each pair has 4 componets: a, b, offset,
            #                                           bond_order, bond_offset
            # a, b: atom index to draw bond
            # offset: original meaning to make offset for mid-point.
            # bond_oder: if not supplied, set it to 1 (single bond).
            #            It can be  1, 2, 3, corresponding to single,
            #            double, triple bond
            # bond_offset: displacement from original bond position.
            #              Default is (bondlinewidth, bondlinewidth, 0)
            #              for bond_order > 1.
            if len(pair) == 2:
                a, b = pair
                offset = (0, 0, 0)
                bond_order = 1
                bond_offset = (0, 0, 0)
            elif len(pair) == 3:
                a, b, offset = pair
                bond_order = 1
                bond_offset = (0, 0, 0)
            elif len(pair) == 4:
                a, b, offset, bond_order = pair
                bond_offset = (self.bondlinewidth, self.bondlinewidth, 0)
            elif len(pair) > 4:
                a, b, offset, bond_order, bond_offset = pair
            else:
                raise RuntimeError('Each list in bondatom must have at least '
                                   '2 entries. Error at %s' % pair)

            if len(offset) != 3:
                raise ValueError('offset must have 3 elements. '
                                 'Error at %s' % pair)
            if len(bond_offset) != 3:
                raise ValueError('bond_offset must have 3 elements. '
                                 'Error at %s' % pair)
            if bond_order not in [0, 1, 2, 3]:
                raise ValueError('bond_order must be either 0, 1, 2, or 3. '
                                 'Error at %s' % pair)

            # Up to here, we should have all a, b, offset, bond_order,
            # bond_offset for all bonds.

            # Rotate bond_offset so that its direction is 90 deg. off the bond
            # Utilize Atoms object to rotate
            if bond_order > 1 and np.linalg.norm(bond_offset) > 1.e-9:
                tmp_atoms = Atoms('H3')
                tmp_atoms.set_cell(self.cell)
                tmp_atoms.set_positions([
                    self.positions[a],
                    self.positions[b],
                    self.positions[b] + np.array(bond_offset),
                ])
                tmp_atoms.center()
                tmp_atoms.set_angle(0, 1, 2, 90)
                bond_offset = tmp_atoms[2].position - tmp_atoms[1].position

            R = np.dot(offset, self.cell)
            mida = 0.5 * (self.positions[a] + self.positions[b] + R)
            midb = 0.5 * (self.positions[a] + self.positions[b] - R)
            if self.textures is not None:
                texa = self.textures[a]
                texb = self.textures[b]
            else:
                texa = texb = 'ase3'

            if self.transmittances is not None:
                transa = self.transmittances[a]
                transb = self.transmittances[b]
            else:
                transa = transb = 0.

            # draw bond, according to its bond_order.
            # bond_order == 0: No bond is plotted
            # bond_order == 1: use original code
            # bond_order == 2: draw two bonds, one is shifted by bond_offset/2,
            #                  and another is shifted by -bond_offset/2.
            # bond_order == 3: draw two bonds, one is shifted by bond_offset,
            #                  and one is shifted by -bond_offset, and the
            #                  other has no shift.
            # To shift the bond, add the shift to the first two coordinate in
            # write statement.

            posa = self.positions[a]
            posb = self.positions[b]
            cola = self.colors[a]
            colb = self.colors[b]

            if bond_order == 1:
                draw_tuples = (
                    (posa, mida, cola, transa, texa),
                    (posb, midb, colb, transb, texb))

            elif bond_order == 2:
                bs = [x / 2 for x in bond_offset]
                draw_tuples = (
                    (posa - bs, mida - bs, cola, transa, texa),
                    (posb - bs, midb - bs, colb, transb, texb),
                    (posa + bs, mida + bs, cola, transa, texa),
                    (posb + bs, midb + bs, colb, transb, texb))

            elif bond_order == 3:
                bs = bond_offset
                draw_tuples = (
                    (posa, mida, cola, transa, texa),
                    (posb, midb, colb, transb, texb),
                    (posa + bs, mida + bs, cola, transa, texa),
                    (posb + bs, midb + bs, colb, transb, texb),
                    (posa - bs, mida - bs, cola, transa, texa),
                    (posb - bs, midb - bs, colb, transb, texb))

            bondatoms += ''.join(f'cylinder {{{pa(p)}, '
                                 f'{pa(m)}, Rbond texture{{pigment '
                                 f'{{color {pc(c)} '
                                 f'transmit {tr}}} finish{{{tx}}}}}}}\n'
                                 for p, m, c, tr, tx in
                                 draw_tuples)

        bondatoms = bondatoms.strip('\n')

        # Draw constraints if requested
        constraints = ''
        if self.exportconstraints:
            for a in self.constrainatoms:
                dia = self.diameters[a]
                loc = self.positions[a]
                trans = 0.0
                if self.transmittances is not None:
                    trans = self.transmittances[a]
                constraints += f'constrain({pa(loc)}, {dia / 2.:.2f}, Black, '\
                    f'{trans}, {tex}) // #{a:n} \n'
        constraints = constraints.strip('\n')

        pov = f"""#version 3.6;
#include "colors.inc"
#include "finish.inc"

global_settings {{assumed_gamma 2.2 max_trace_level 6}}
background {{{pc(self.background)}{' transmit 1.0' if self.transparent else ''}}}
camera {{{self.camera_type}
  right -{self.image_width:.2f}*x up {self.image_height:.2f}*y
  direction {self.image_plane:.2f}*z
  location <0,0,{self.camera_dist:.2f}> look_at <0,0,0>}}
{point_lights}
{area_light if area_light != '' else '// no area light'}
{fog if fog != '' else '// no fog'}
{mat_style_keys}
#declare Rcell = {self.celllinewidth:.3f};
#declare Rbond = {self.bondlinewidth:.3f};

#macro atom(LOC, R, COL, TRANS, FIN)
  sphere{{LOC, R texture{{pigment{{color COL transmit TRANS}} finish{{FIN}}}}}}
#end
#macro constrain(LOC, R, COL, TRANS FIN)
union{{torus{{R, Rcell rotate 45*z texture{{pigment{{color COL transmit TRANS}} finish{{FIN}}}}}}
     torus{{R, Rcell rotate -45*z texture{{pigment{{color COL transmit TRANS}} finish{{FIN}}}}}}
     translate LOC}}
#end

{cell_vertices if cell_vertices != '' else '// no cell vertices'}
{atoms}
{bondatoms}
{constraints if constraints != '' else '// no constraints'}
"""  # noqa: E501

        with open(path, 'w') as fd:
            fd.write(pov)

        return path

    def write(self, pov_path):
        pov_path = require_pov(pov_path)
        ini_path = pov_path.with_suffix('.ini')
        self.write_ini(ini_path)
        self.write_pov(pov_path)
        if self.isosurfaces is not None:
            with open(pov_path, 'a') as fd:
                for iso in self.isosurfaces:
                    fd.write(iso.format_mesh())
        return POVRAYInputs(ini_path)


def require_pov(path):
    path = Path(path)
    if path.suffix != '.pov':
        raise ValueError(f'Expected .pov path, got {path}')
    return path


class POVRAYInputs:
    def __init__(self, path):
        self.path = path

    def render(self, povray_executable='povray', stderr=DEVNULL,
               clean_up=False):
        cmd = [povray_executable, str(self.path)]

        check_call(cmd, stderr=stderr)
        png_path = self.path.with_suffix('.png').absolute()
        if not png_path.is_file():
            raise RuntimeError(f'Povray left no output PNG file "{png_path}"')

        if clean_up:
            unlink(self.path)
            unlink(self.path.with_suffix('.pov'))

        return png_path


class POVRAYIsosurface:
    def __init__(self, density_grid, cut_off, cell, cell_origin,
                 closed_edges=False, gradient_ascending=False,
                 color=(0.85, 0.80, 0.25, 0.2), material='ase3'):
        """
        density_grid: 3D float ndarray
            A regular grid on that spans the cell. The first dimension
            corresponds to the first cell vector and so on.
        cut_off: float
            The density value of the isosurface.
        cell: 2D float ndarray or ASE cell object
            The 3 vectors which give the cell's repetition
        cell_origin: 4 float tuple
            The cell origin as used by POVRAY object
        closed_edges: bool
            Setting this will fill in isosurface edges at the cell boundaries.
            Filling in the edges can help with visualizing
            highly porous structures.
        gradient_ascending: bool
            Lets you pick the area you want to enclose, i.e., should the denser
            or less dense area be filled in.
        color: povray color string, float, or float tuple
            1 float is interpreted as grey scale, a 3 float tuple is rgb,
            4 float tuple is rgbt, and 5 float tuple is rgbft, where
            t is transmission fraction and f is filter fraction.
            Named Povray colors are set in colors.inc
            (http://wiki.povray.org/content/Reference:Colors.inc)
        material: string
            Can be a finish macro defined by POVRAY.material_styles
            or a full Povray material {...} specification. Using a
            full material specification willoverride the color parameter.
        """

        self.gradient_direction = 'ascent' if gradient_ascending else 'descent'
        self.color = color
        self.material = material
        self.closed_edges = closed_edges
        self._cut_off = cut_off

        if self.gradient_direction == 'ascent':
            cv = 2 * cut_off
        else:
            cv = 0

        if closed_edges:
            shape_old = density_grid.shape
            # since well be padding, we need to keep the data at origin
            cell_origin += -(1.0 / np.array(shape_old)) @ cell
            density_grid = np.pad(
                density_grid, pad_width=(
                    1,), mode='constant', constant_values=cv)
            shape_new = density_grid.shape
            s = np.array(shape_new) / np.array(shape_old)
            cell = cell @ np.diag(s)

        self.cell = cell
        self.cell_origin = cell_origin
        self.density_grid = density_grid
        self.spacing = tuple(1.0 / np.array(self.density_grid.shape))

        scaled_verts, faces, _normals, _values = self.compute_mesh(
            self.density_grid,
            self.cut_off,
            self.spacing,
            self.gradient_direction)

        # The verts are scaled by default, this is the super easy way of
        # distributing them in real space but it's easier to do affine
        # transformations/rotations on a unit cube so I leave it like that
        # verts = scaled_verts.dot(self.cell)
        self.verts = scaled_verts
        self.faces = faces

    @property
    def cut_off(self):
        return self._cut_off

    @cut_off.setter
    def cut_off(self, value):
        raise Exception("Use the set_cut_off method")

    def set_cut_off(self, value):
        self._cut_off = value

        if self.gradient_direction == 'ascent':
            cv = 2 * self.cut_off
        else:
            cv = 0

        if self.closed_edges:
            shape_old = self.density_grid.shape
            # since well be padding, we need to keep the data at origin
            self.cell_origin += -(1.0 / np.array(shape_old)) @ self.cell
            self.density_grid = np.pad(
                self.density_grid, pad_width=(
                    1,), mode='constant', constant_values=cv)
            shape_new = self.density_grid.shape
            s = np.array(shape_new) / np.array(shape_old)
            self.cell = self.cell @ np.diag(s)

        self.spacing = tuple(1.0 / np.array(self.density_grid.shape))

        scaled_verts, faces, _, _ = self.compute_mesh(
            self.density_grid,
            self.cut_off,
            self.spacing,
            self.gradient_direction)

        self.verts = scaled_verts
        self.faces = faces

    @classmethod
    def from_POVRAY(cls, povray, density_grid, cut_off, **kwargs):
        return cls(cell=povray.cell,
                   cell_origin=povray.cell_vertices[0, 0, 0],
                   density_grid=density_grid,
                   cut_off=cut_off, **kwargs)

    @staticmethod
    def wrapped_triples_section(triple_list,
                                triple_format="<{:f}, {:f}, {:f}>".format,
                                triples_per_line=4):

        triples = [triple_format(*x) for x in triple_list]
        n = len(triples)
        s = ''
        tpl = triples_per_line
        c = 0

        while c < n - tpl:
            c += tpl
            s += '\n     '
            s += ', '.join(triples[c - tpl:c])
        s += '\n    '
        s += ', '.join(triples[c:])
        return s

    @staticmethod
    def compute_mesh(density_grid, cut_off, spacing, gradient_direction):
        """

        Import statement is in this method and not file header
        since few users will use isosurface rendering.

        Returns scaled_verts, faces, normals, values. See skimage docs.

        """

        # marching_cubes name was changed in skimage v0.19
        try:
            # New skimage
            from skimage.measure import marching_cubes
        except ImportError:
            # Old skimage (remove at some point)
            from skimage.measure import marching_cubes_lewiner as marching_cubes

        return marching_cubes(
            density_grid,
            level=cut_off,
            spacing=spacing,
            gradient_direction=gradient_direction,
            allow_degenerate=False)

    def format_mesh(self):
        """Returns a formatted data output for POVRAY files

        Example:
        material = '''
          material { // This material looks like pink jelly
            texture {
              pigment { rgbt <0.8, 0.25, 0.25, 0.5> }
              finish{ diffuse 0.85 ambient 0.99 brilliance 3 specular 0.5 roughness 0.001
                reflection { 0.05, 0.98 fresnel on exponent 1.5 }
                conserve_energy
              }
            }
            interior { ior 1.3 }
          }
          photons {
              target
              refraction on
              reflection on
              collect on
          }'''
        """  # noqa: E501

        if self.material in POVRAY.material_styles_dict:
            material = f"""material {{
        texture {{
          pigment {{ {pc(self.color)} }}
          finish {{ {self.material} }}
        }}
      }}"""
        else:
            material = self.material

        # Start writing the mesh2
        vertex_vectors = self.wrapped_triples_section(
            triple_list=self.verts,
            triple_format="<{:f}, {:f}, {:f}>".format,
            triples_per_line=4)

        face_indices = self.wrapped_triples_section(
            triple_list=self.faces,
            triple_format="<{:n}, {:n}, {:n}>".format,
            triples_per_line=5)

        cell = self.cell
        cell_or = self.cell_origin
        mesh2 = f"""\n\nmesh2 {{
    vertex_vectors {{  {len(self.verts):n},
    {vertex_vectors}
    }}
    face_indices {{ {len(self.faces):n},
    {face_indices}
    }}
{material if material != '' else '// no material'}
  matrix < {cell[0][0]:f}, {cell[0][1]:f}, {cell[0][2]:f},
           {cell[1][0]:f}, {cell[1][1]:f}, {cell[1][2]:f},
           {cell[2][0]:f}, {cell[2][1]:f}, {cell[2][2]:f},
           {cell_or[0]:f}, {cell_or[1]:f}, {cell_or[2]:f}>
    }}
    """
        return mesh2


def pop_deprecated(dct, name):
    import warnings
    if name in dct:
        del dct[name]
        warnings.warn(f'The "{name}" keyword of write_pov() is deprecated '
                      'and has no effect; this will raise an error in the '
                      'future.', FutureWarning)


def write_pov(filename, atoms, *,
              povray_settings=None, isosurface_data=None,
              **generic_projection_settings):

    for name in ['run_povray', 'povray_path', 'stderr', 'extras']:
        pop_deprecated(generic_projection_settings, name)

    if povray_settings is None:
        povray_settings = {}

    pvars = PlottingVariables(atoms, scale=1.0, **generic_projection_settings)
    pov_obj = POVRAY.from_PlottingVariables(pvars, **povray_settings)

    if isosurface_data is None:
        isosurface_data = []
    elif not isinstance(isosurface_data, Sequence):
        isosurface_data = [isosurface_data]

    isosurfaces = []
    for isodata in isosurface_data:
        if isinstance(isodata, POVRAYIsosurface):
            iso = isodata
        else:
            iso = POVRAYIsosurface.from_POVRAY(pov_obj, **isodata)
        isosurfaces.append(iso)
    pov_obj.isosurfaces = isosurfaces

    return pov_obj.write(filename)



from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack
from ase.units import Bohr, Hartree, fs, kB, kcal, kJ, mol


def test_things():

    assert [0, 0, 0] in monkhorst_pack((1, 3, 5)).tolist()
    assert [0, 0, 0] not in monkhorst_pack((1, 3, 6)).tolist()
    assert len(monkhorst_pack((3, 4, 6))) == 3 * 4 * 6

    print(Hartree, Bohr, kJ / mol, kcal / mol, kB * 300, fs, 1 / fs)

    hcp = bulk('X', 'hcp', a=1) * (2, 2, 1)
    assert abs(hcp.get_distance(0, 3, mic=True) - 1) < 1e-12
    assert abs(hcp.get_distance(0, 4, mic=True) - 1) < 1e-12
    assert abs(hcp.get_distance(2, 5, mic=True) - 1) < 1e-12



"""
Routines for manually parsing a lammps data file.  This is a simplified
recreation of ase.io.lammpsdata's read functionality that we use for
testing so as to attempt to provide an independent verification.  It is
fairly ad hoc and was designed specifically to work for the
'lammpsdata_input.data' file under ase.test.testdata.  In particular,
the following assumptions are made about the lammps data file:

  - Only a single atomic species is present

  - The cell is orthogonal (xy, xz, yz tilt factors are
    ignored even if they exist)
"""
import io
import pathlib
import re

import numpy as np

from ase.calculators.lammps import convert


def split_contents_by_section(raw_datafile_contents):
    # Remove comments
    raw_datafile_contents = re.sub('#.*', '', raw_datafile_contents)
    return re.split(r'^([A-Za-z]+\s*)$\n',
                    raw_datafile_contents, flags=re.MULTILINE)


def extract_section(raw_datafile_contents, section_header):

    contents_split_by_section = split_contents_by_section(raw_datafile_contents)

    section = None

    for ind, block in enumerate(contents_split_by_section):
        if block.startswith(section_header):
            section = contents_split_by_section[ind + 1].strip()
            break

    return section


def extract_cell(raw_datafile_contents):
    """
    NOTE: Assumes an orthogonal cell (xy, xz, yz tilt factors are
    ignored even if they exist)
    """
    RE_CELL = re.compile(
        r'''
            (\S+)\s+(\S+)\s+xlo\s+xhi\n
            (\S+)\s+(\S+)\s+ylo\s+yhi\n
            (\S+)\s+(\S+)\s+zlo\s+zhi\n
        ''',
        flags=re.VERBOSE,
    )
    xlo, xhi, ylo, yhi, zlo, zhi = map(
        float, RE_CELL.search(raw_datafile_contents).groups()
    )

    cell = [[xhi - xlo, 0, 0], [0, yhi - ylo, 0], [0, 0, zhi - zlo]]

    return cell


def extract_mass(raw_datafile_contents):
    """
    NOTE: Assumes that only a single atomic species is present
    """
    masses_block = extract_section(raw_datafile_contents, 'Masses')

    if masses_block is None:
        return None
    else:
        mass = re.match(r'\s*[0-9]+\s+(\S+)', masses_block).group(1)
        return float(mass)


def extract_atom_quantities(raw_datafile_contents):

    # Grab all atoms lines
    atoms_block = extract_section(raw_datafile_contents, 'Atoms')

    # Now parse each individual atoms line for quantities
    charges = []
    positions = []
    travels = []

    RE_ATOM_LINE = re.compile(
        r'\s*[0-9]+\s+[0-9]+\s+[0-9]+\s+(\S+)\s+'
        r'(\S+)\s+(\S+)\s+(\S+)\s?'
        r'([0-9-]+)?\s?([0-9-]+)?\s?([0-9-]+)?'
    )

    for atom_line in atoms_block.splitlines():
        q, x, y, z, *travel = RE_ATOM_LINE.match(atom_line).groups()
        charges.append(float(q))
        positions.append(list(map(float, [x, y, z])))
        if None not in travel:
            travels.append(list(map(int, travel)))
        else:
            travels.append(None)

    return charges, positions, travels


def extract_velocities(raw_datafile_contents):
    """
    NOTE: Assumes metal units are used in data file
    """
    velocities_block = extract_section(raw_datafile_contents, 'Velocities')

    RE_VELOCITY = re.compile(
        r'\s*[0-9]+\s+(\S+)\s+(\S+)\s+(\S+)'
    )

    # Now parse each individual line for velocity
    velocities = []
    for velocities_line in velocities_block.splitlines():
        v = RE_VELOCITY.match(velocities_line).groups()
        velocities.append(list(map(float, v)))

    # Convert to ASE's velocity units (uses unusual unit of time)
    velocities = convert(np.asarray(velocities), 'velocity', 'metal', 'ASE')

    return velocities


def lammpsdata_file_extracted_sections(lammpsdata):
    """
    Manually read a lammpsdata file and grep for the different
    quantities we want to check.  Accepts either a string indicating the name
    of the file, a pathlib.Path object indicating the location of the file, a
    StringIO object containing the file contents, or a file object
    """
    if isinstance(lammpsdata, str) or isinstance(lammpsdata, pathlib.Path):
        with open(lammpsdata) as fd:
            raw_datafile_contents = fd.read()

    elif isinstance(lammpsdata, io.StringIO):
        raw_datafile_contents = lammpsdata.getvalue()

    elif isinstance(lammpsdata, io.TextIOBase):
        raw_datafile_contents = lammpsdata.read()

    else:
        raise ValueError(
            'Lammps data file content inputted in unsupported '
            'object type {type(lammpsdata)}'
        )

    cell = extract_cell(raw_datafile_contents)
    mass = extract_mass(raw_datafile_contents)
    charges, positions, travels = extract_atom_quantities(raw_datafile_contents)
    velocities = extract_velocities(raw_datafile_contents)

    return {
        'cell': cell,
        'mass': mass,
        'charges': charges,
        'positions': positions,
        'travels': travels,
        'velocities': velocities,
    }



import gzip
import struct
from collections import deque
from os.path import splitext

import numpy as np

from ase.atoms import Atoms
from ase.calculators.lammps import convert
from ase.calculators.singlepoint import SinglePointCalculator
from ase.parallel import paropen
from ase.quaternions import Quaternions


def read_lammps_dump(infileobj, **kwargs):
    """Method which reads a LAMMPS dump file.

       LAMMPS chooses output method depending on the given suffix:
        - .bin  : binary file
        - .gz   : output piped through gzip
        - .mpiio: using mpiio (should be like cleartext,
                  with different ordering)
        - else  : normal clear-text format

    :param infileobj: string to file, opened file or file-like stream

    """
    # !TODO: add support for lammps-regex naming schemes (output per
    # processor and timestep wildcards)

    opened = False
    if isinstance(infileobj, str):
        opened = True
        suffix = splitext(infileobj)[-1]
        if suffix == ".bin":
            fileobj = paropen(infileobj, "rb")
        elif suffix == ".gz":
            # !TODO: save for parallel execution?
            fileobj = gzip.open(infileobj, "rb")
        else:
            fileobj = paropen(infileobj)
    else:
        suffix = splitext(infileobj.name)[-1]
        fileobj = infileobj

    if suffix == ".bin":
        out = read_lammps_dump_binary(fileobj, **kwargs)
        if opened:
            fileobj.close()
        return out

    out = read_lammps_dump_text(fileobj, **kwargs)

    if opened:
        fileobj.close()

    return out


def lammps_data_to_ase_atoms(
    data,
    colnames,
    cell,
    celldisp,
    pbc=False,
    atomsobj=Atoms,
    order=True,
    specorder=None,
    prismobj=None,
    units="metal",
):
    """Extract positions and other per-atom parameters and create Atoms

    :param data: per atom data
    :param colnames: index for data
    :param cell: cell dimensions
    :param celldisp: origin shift
    :param pbc: periodic boundaries
    :param atomsobj: function to create ase-Atoms object
    :param order: sort atoms by id. Might be faster to turn off.
    Disregarded in case `id` column is not given in file.
    :param specorder: list of species to map lammps types to ase-species
    (usually .dump files to not contain type to species mapping)
    :param prismobj: Coordinate transformation between lammps and ase
    :type prismobj: Prism
    :param units: lammps units for unit transformation between lammps and ase
    :returns: Atoms object
    :rtype: Atoms

    """
    if len(data.shape) == 1:
        data = data[np.newaxis, :]

    # read IDs if given and order if needed
    if "id" in colnames:
        ids = data[:, colnames.index("id")].astype(int)
        if order:
            sort_order = np.argsort(ids)
            data = data[sort_order, :]

    # determine the elements
    if "element" in colnames:
        # priority to elements written in file
        elements = data[:, colnames.index("element")]
    elif "type" in colnames:
        # fall back to `types` otherwise
        elements = data[:, colnames.index("type")].astype(int)

        # reconstruct types from given specorder
        if specorder:
            elements = [specorder[t - 1] for t in elements]
    else:
        # todo: what if specorder give but no types?
        # in principle the masses could work for atoms, but that needs
        # lots of cases and new code I guess
        raise ValueError("Cannot determine atom types form LAMMPS dump file")

    def get_quantity(labels, quantity=None):
        try:
            cols = [colnames.index(label) for label in labels]
            if quantity:
                return convert(data[:, cols].astype(float), quantity,
                               units, "ASE")

            return data[:, cols].astype(float)
        except ValueError:
            return None

    # Positions
    positions = None
    scaled_positions = None
    if "x" in colnames:
        # doc: x, y, z = unscaled atom coordinates
        positions = get_quantity(["x", "y", "z"], "distance")
    elif "xs" in colnames:
        # doc: xs,ys,zs = scaled atom coordinates
        scaled_positions = get_quantity(["xs", "ys", "zs"])
    elif "xu" in colnames:
        # doc: xu,yu,zu = unwrapped atom coordinates
        positions = get_quantity(["xu", "yu", "zu"], "distance")
    elif "xsu" in colnames:
        # xsu,ysu,zsu = scaled unwrapped atom coordinates
        scaled_positions = get_quantity(["xsu", "ysu", "zsu"])
    else:
        raise ValueError("No atomic positions found in LAMMPS output")

    velocities = get_quantity(["vx", "vy", "vz"], "velocity")
    charges = get_quantity(["q"], "charge")
    forces = get_quantity(["fx", "fy", "fz"], "force")
    # !TODO: how need quaternions be converted?
    quaternions = get_quantity(["c_q[1]", "c_q[2]", "c_q[3]", "c_q[4]"])

    # convert cell
    cell = convert(cell, "distance", units, "ASE")
    celldisp = convert(celldisp, "distance", units, "ASE")
    if prismobj:
        celldisp = prismobj.vector_to_ase(celldisp)
        cell = prismobj.update_cell(cell)

    if quaternions is not None:
        out_atoms = Quaternions(
            symbols=elements,
            positions=positions,
            cell=cell,
            celldisp=celldisp,
            pbc=pbc,
            quaternions=quaternions,
        )
    elif positions is not None:
        # reverse coordinations transform to lammps system
        # (for all vectors = pos, vel, force)
        if prismobj:
            positions = prismobj.vector_to_ase(positions, wrap=True)

        out_atoms = atomsobj(
            symbols=elements,
            positions=positions,
            pbc=pbc,
            celldisp=celldisp,
            cell=cell
        )
    elif scaled_positions is not None:
        out_atoms = atomsobj(
            symbols=elements,
            scaled_positions=scaled_positions,
            pbc=pbc,
            celldisp=celldisp,
            cell=cell,
        )

    if velocities is not None:
        if prismobj:
            velocities = prismobj.vector_to_ase(velocities)
        out_atoms.set_velocities(velocities)
    if charges is not None:
        out_atoms.set_initial_charges([charge[0] for charge in charges])
    if forces is not None:
        if prismobj:
            forces = prismobj.vector_to_ase(forces)
        # !TODO: use another calculator if available (or move forces
        #        to atoms.property) (other problem: synchronizing
        #        parallel runs)
        calculator = SinglePointCalculator(out_atoms, energy=0.0,
                                           forces=forces)
        out_atoms.calc = calculator

    # process the extra columns of fixes, variables and computes
    #    that can be dumped, add as additional arrays to atoms object
    for colname in colnames:
        # determine if it is a compute or fix (but not the quaternian)
        if (colname.startswith('f_') or colname.startswith('v_') or
                (colname.startswith('c_') and not colname.startswith('c_q['))):
            out_atoms.new_array(colname, get_quantity([colname]),
                                dtype='float')

    return out_atoms


def construct_cell(diagdisp, offdiag):
    """Help function to create an ASE-cell with displacement vector from
    the lammps coordination system parameters.

    :param diagdisp: cell dimension convoluted with the displacement vector
    :param offdiag: off-diagonal cell elements
    :returns: cell and cell displacement vector
    :rtype: tuple
    """
    xlo, xhi, ylo, yhi, zlo, zhi = diagdisp
    xy, xz, yz = offdiag

    # create ase-cell from lammps-box
    xhilo = (xhi - xlo) - abs(xy) - abs(xz)
    yhilo = (yhi - ylo) - abs(yz)
    zhilo = zhi - zlo
    celldispx = xlo - min(0, xy) - min(0, xz)
    celldispy = ylo - min(0, yz)
    celldispz = zlo
    cell = np.array([[xhilo, 0, 0], [xy, yhilo, 0], [xz, yz, zhilo]])
    celldisp = np.array([celldispx, celldispy, celldispz])

    return cell, celldisp


def get_max_index(index):
    if np.isscalar(index):
        return index
    elif isinstance(index, slice):
        return index.stop if (index.stop is not None) else float("inf")


def read_lammps_dump_text(fileobj, index=-1, **kwargs):
    """Process cleartext lammps dumpfiles

    :param fileobj: filestream providing the trajectory data
    :param index: integer or slice object (default: get the last timestep)
    :returns: list of Atoms objects
    :rtype: list
    """
    # Load all dumped timesteps into memory simultaneously
    lines = deque(fileobj.readlines())
    index_end = get_max_index(index)

    n_atoms = 0
    images = []

    # avoid references before assignment in case of incorrect file structure
    cell, celldisp, pbc = None, None, False

    while len(lines) > n_atoms:
        line = lines.popleft()

        if "ITEM: TIMESTEP" in line:
            n_atoms = 0
            line = lines.popleft()
            # !TODO: pyflakes complains about this line -> do something
            # ntimestep = int(line.split()[0])  # NOQA

        if "ITEM: NUMBER OF ATOMS" in line:
            line = lines.popleft()
            n_atoms = int(line.split()[0])

        if "ITEM: BOX BOUNDS" in line:
            # save labels behind "ITEM: BOX BOUNDS" in triclinic case
            # (>=lammps-7Jul09)
            tilt_items = line.split()[3:]
            celldatarows = [lines.popleft() for _ in range(3)]
            celldata = np.loadtxt(celldatarows)
            diagdisp = celldata[:, :2].reshape(6, 1).flatten()

            # determine cell tilt (triclinic case!)
            if len(celldata[0]) > 2:
                # for >=lammps-7Jul09 use labels behind "ITEM: BOX BOUNDS"
                # to assign tilt (vector) elements ...
                offdiag = celldata[:, 2]
                # ... otherwise assume default order in 3rd column
                # (if the latter was present)
                if len(tilt_items) >= 3:
                    sort_index = [tilt_items.index(i)
                                  for i in ["xy", "xz", "yz"]]
                    offdiag = offdiag[sort_index]
            else:
                offdiag = (0.0,) * 3

            cell, celldisp = construct_cell(diagdisp, offdiag)

            # Handle pbc conditions
            if len(tilt_items) == 3:
                pbc_items = tilt_items
            elif len(tilt_items) > 3:
                pbc_items = tilt_items[3:6]
            else:
                pbc_items = ["f", "f", "f"]
            pbc = ["p" in d.lower() for d in pbc_items]

        if "ITEM: ATOMS" in line:
            colnames = line.split()[2:]
            datarows = [lines.popleft() for _ in range(n_atoms)]
            data = np.loadtxt(datarows, dtype=str, ndmin=2)
            out_atoms = lammps_data_to_ase_atoms(
                data=data,
                colnames=colnames,
                cell=cell,
                celldisp=celldisp,
                atomsobj=Atoms,
                pbc=pbc,
                **kwargs
            )
            images.append(out_atoms)

        if len(images) > index_end >= 0:
            break

    return images[index]


def read_lammps_dump_binary(
    fileobj, index=-1, colnames=None, intformat="SMALLBIG", **kwargs
):
    """Read binary dump-files (after binary2txt.cpp from lammps/tools)

    :param fileobj: file-stream containing the binary lammps data
    :param index: integer or slice object (default: get the last timestep)
    :param colnames: data is columns and identified by a header
    :param intformat: lammps support different integer size.  Parameter set \
    at compile-time and can unfortunately not derived from data file
    :returns: list of Atoms-objects
    :rtype: list
    """
    # depending on the chosen compilation flag lammps uses either normal
    # integers or long long for its id or timestep numbering
    # !TODO: tags are cast to double -> missing/double ids (add check?)
    _tagformat, bigformat = dict(
        SMALLSMALL=("i", "i"), SMALLBIG=("i", "q"), BIGBIG=("q", "q")
    )[intformat]

    index_end = get_max_index(index)

    # Standard columns layout from lammpsrun
    if not colnames:
        colnames = ["id", "type", "x", "y", "z",
                    "vx", "vy", "vz", "fx", "fy", "fz"]

    images = []

    # wrap struct.unpack to raise EOFError
    def read_variables(string):
        obj_len = struct.calcsize(string)
        data_obj = fileobj.read(obj_len)
        if obj_len != len(data_obj):
            raise EOFError
        return struct.unpack(string, data_obj)

    while True:
        try:
            # Assume that the binary dump file is in the old (pre-29Oct2020)
            # format
            magic_string = None

            # read header
            ntimestep, = read_variables("=" + bigformat)

            # In the new LAMMPS binary dump format (version 29Oct2020 and
            # onward), a negative timestep is used to indicate that the next
            # few bytes will contain certain metadata
            if ntimestep < 0:
                # First bigint was actually encoding the negative of the format
                # name string length (we call this 'magic_string' to
                magic_string_len = -ntimestep

                # The next `magic_string_len` bytes will hold a string
                # indicating the format of the dump file
                magic_string = b''.join(read_variables(
                    "=" + str(magic_string_len) + "c"))

                # Read endianness (integer). For now, we'll disregard the value
                # and simply use the host machine's endianness (via '='
                # character used with struct.calcsize).
                #
                # TODO: Use the endianness of the dump file in subsequent
                #       read_variables rather than just assuming it will match
                #       that of the host
                read_variables("=i")

                # Read revision number (integer)
                revision, = read_variables("=i")

                # Finally, read the actual timestep (bigint)
                ntimestep, = read_variables("=" + bigformat)

            _n_atoms, triclinic = read_variables("=" + bigformat + "i")
            boundary = read_variables("=6i")
            diagdisp = read_variables("=6d")
            if triclinic != 0:
                offdiag = read_variables("=3d")
            else:
                offdiag = (0.0,) * 3
            size_one, = read_variables("=i")

            if len(colnames) != size_one:
                raise ValueError("Provided columns do not match binary file")

            if magic_string and revision > 1:
                # New binary dump format includes units string,
                # columns string, and time
                units_str_len, = read_variables("=i")

                if units_str_len > 0:
                    # Read lammps units style
                    _ = b''.join(
                        read_variables("=" + str(units_str_len) + "c"))

                flag, = read_variables("=c")
                if flag != b'\x00':
                    # Flag was non-empty string
                    read_variables("=d")

                # Length of column string
                columns_str_len, = read_variables("=i")

                # Read column string (e.g., "id type x y z vx vy vz fx fy fz")
                _ = b''.join(read_variables("=" + str(columns_str_len) + "c"))

            nchunk, = read_variables("=i")

            # lammps cells/boxes can have different boundary conditions on each
            # sides (makes mainly sense for different non-periodic conditions
            # (e.g. [f]ixed and [s]hrink for a irradiation simulation))
            # periodic case: b 0 = 'p'
            # non-peridic cases 1: 'f', 2 : 's', 3: 'm'
            pbc = np.sum(np.array(boundary).reshape((3, 2)), axis=1) == 0

            cell, celldisp = construct_cell(diagdisp, offdiag)

            data = []
            for _ in range(nchunk):
                # number-of-data-entries
                n_data, = read_variables("=i")
                # retrieve per atom data
                data += read_variables("=" + str(n_data) + "d")
            data = np.array(data).reshape((-1, size_one))

            # map data-chunk to ase atoms
            out_atoms = lammps_data_to_ase_atoms(
                data=data,
                colnames=colnames,
                cell=cell,
                celldisp=celldisp,
                pbc=pbc,
                **kwargs
            )

            images.append(out_atoms)

            # stop if requested index has been found
            if len(images) > index_end >= 0:
                break

        except EOFError:
            break

    return images[index]



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from scipy.integrate import trapezoid

from ase import Atoms
from ase.calculators.calculator import CalculationFailed, CalculatorSetupError
from ase.calculators.emt import EMT
from ase.calculators.harmonic import HarmonicCalculator, HarmonicForceField
from ase.calculators.mixing import MixedCalculator
from ase.geometry.geometry import get_distances_derivatives
from ase.md.andersen import Andersen
from ase.md.velocitydistribution import (
    MaxwellBoltzmannDistribution,
    Stationary,
    ZeroRotation,
)
from ase.optimize import BFGS
from ase.units import fs
from ase.vibrations import Vibrations

ref_pos = np.asarray([[8.7161, 7.96276, 8.48206], [8.60594, 8.04985, 9.44464],
                      [8.0154, 8.52264, 8.10545]])
ref_atoms = Atoms('OH2', positions=ref_pos)  # relaxed water molecule
ref_energy = -14.222189  # y shift of the 'parabola' (harmonic potential)

# example Hessian matrix as obtained from DFT
hessian_x = np.asarray([[2.82630333e+01, -2.24763667e+01, 7.22478333e+00,
                         -2.96970000e+00, 2.34363333e+00, 2.72788333e+00,
                         -2.52159833e+01, 2.01307833e+01, -9.94651667e+00],
                        [-2.24763667e+01, 1.78621333e+01, -5.77378333e+00,
                         2.33703333e+00, -1.85276667e+00, -2.15118333e+00,
                         2.01258667e+01, -1.60350833e+01, 7.93248333e+00],
                        [7.22478333e+00, -5.77378333e+00, 5.72735000e+01,
                         7.25470000e+00, -5.75313333e+00, -4.69477333e+01,
                         -1.44613000e+01, 1.15504833e+01, -1.03523333e+01],
                        [-2.96970000e+00, 2.33703333e+00, 7.25470000e+00,
                         2.96170000e+00, -2.36901667e+00, -3.76841667e+00,
                         -2.83833333e-02, 3.06833333e-02, -3.49190000e+00],
                        [2.34363333e+00, -1.85276667e+00, -5.75313333e+00,
                         -2.36901667e+00, 1.89046667e+00, 2.95495000e+00,
                         2.90666667e-02, -1.80666667e-02, 2.79565000e+00],
                        [2.72788333e+00, -2.15118333e+00, -4.69477333e+01,
                         -3.76841667e+00, 2.95495000e+00, 4.89340000e+01,
                         1.03146667e+00, -8.18450000e-01, -1.96118333e+00],
                        [-2.52159833e+01, 2.01258667e+01, -1.44613000e+01,
                         -2.83833333e-02, 2.90666667e-02, 1.03146667e+00,
                         2.52034000e+01, -2.01516833e+01, 1.34293167e+01],
                        [2.01307833e+01, -1.60350833e+01, 1.15504833e+01,
                         3.06833333e-02, -1.80666667e-02, -8.18450000e-01,
                         -2.01516833e+01, 1.60592333e+01, -1.07369667e+01],
                        [-9.94651667e+00, 7.93248333e+00, -1.03523333e+01,
                         -3.49190000e+00, 2.79565000e+00, -1.96118333e+00,
                         1.34293167e+01, -1.07369667e+01, 1.23150000e+01]])


def assert_water_is_relaxed(atoms):
    forces = atoms.get_forces()
    assert np.allclose(np.zeros(forces.shape), forces)
    assert np.allclose(ref_energy, atoms.get_potential_energy())
    assert np.allclose(atoms.get_angle(1, 0, 2), ref_atoms.get_angle(1, 0, 2))
    assert np.allclose(atoms.get_distance(0, 1), ref_atoms.get_distance(0, 1))
    assert np.allclose(atoms.get_distance(0, 2), ref_atoms.get_distance(0, 2))


def run_optimize(atoms):
    opt = BFGS(atoms)
    opt.run(fmax=1e-9)


@pytest.mark.optimize()
def test_cartesians():
    """In Cartesian coordinates the first 6 trash eigenvalues (translations and
    rotations) can be slightly different from zero; hence set them to zero
    using an increased parameter zero_thresh.
    """
    zero_thresh = 0.06  # set eigvals to zero if abs(eigenvalue) < zero_thresh
    hff = HarmonicForceField(ref_atoms=ref_atoms, ref_energy=ref_energy,
                             hessian_x=hessian_x, zero_thresh=zero_thresh)
    assert np.allclose(hff.hessian_q, hff.hessian_x)
    atoms = ref_atoms.copy()
    atoms.calc = HarmonicCalculator(hff)
    assert_water_is_relaxed(atoms)  # atoms has not been distorted
    run_optimize(atoms)             # nothing should happen
    assert_water_is_relaxed(atoms)  # atoms should still be relaxed
    atoms.set_distance(0, 1, 3.5)   # now distort atoms along axis, no rotation
    run_optimize(atoms)             # optimization should recover original
    assert_water_is_relaxed(atoms)    # relaxed geometry

    with pytest.raises(AssertionError):
        atoms.rattle()                  # relaxation should fail to recover the
        atoms.rotate(90, 'x')           # original geometry of the atoms,
        run_optimize(atoms)             # because Cartesian coordinates are
        assert_water_is_relaxed(atoms)  # not rotationally invariant.


@pytest.mark.optimize()
def test_constraints_with_cartesians():
    """Project out forces along x-component of H-atom (index 0 in the q-vector
    with the Cartesian coordinates (here: x=q)). A change in the x-component of
    the H-atom should not result in restoring forces, when they were projected
    out from the Hessian matrix.
    """
    def test_forces(calc):
        atoms = ref_atoms.copy()
        atoms.calc = calc
        pos = ref_pos.copy()
        pos[0, 0] *= 2
        atoms.set_positions(pos)
        run_optimize(atoms)  # (no) restoring force along distorted x-component
        xdiff = atoms.get_positions() - ref_pos
        return all(xdiff[xdiff != 0] == pos[0, 0] / 2)

    zero_thresh = 0.06  # set eigvals to zero if abs(eigenvalue) < zero_thresh
    parameters = {'ref_atoms': ref_atoms, 'ref_energy': ref_energy,
                  'hessian_x': hessian_x, 'zero_thresh': zero_thresh}
    hff = HarmonicForceField(**parameters)
    calc = HarmonicCalculator(hff)
    assert not test_forces(calc)  # restoring force along distorted x-component

    parameters['constrained_q'] = [0]  # project out the coordinate with index 0
    hff = HarmonicForceField(**parameters)
    calc = HarmonicCalculator(hff)
    assert test_forces(calc)  # no restoring force along distorted x-component


def setup_water(calc):
    atoms = ref_atoms.copy()
    atoms.calc = calc
    assert_water_is_relaxed(atoms)
    atoms.rattle(0.3)
    atoms.rotate(160, 'x')
    assert not np.allclose(ref_energy, atoms.get_potential_energy())
    return atoms


# start doc example 3
dist_defs = [[0, 1], [1, 2], [2, 0]]  # define three distances by atom indices


def water_get_q_from_x(atoms):
    """Simple internal coordinates to describe water with three distances."""
    q_vec = [atoms.get_distance(i, j) for i, j in dist_defs]
    return np.asarray(q_vec)


def water_get_jacobian(atoms):
    """Function to return the Jacobian for the water molecule described by
    three distances."""
    pos = atoms.get_positions()
    dist_vecs = [pos[j] - pos[i] for i, j in dist_defs]
    derivs = get_distances_derivatives(dist_vecs)
    jac = []
    for i, defin in enumerate(dist_defs):
        dqi_dxj = np.zeros(ref_pos.shape)
        for j, deriv in enumerate(derivs[i]):
            dqi_dxj[defin[j]] = deriv
        jac.append(dqi_dxj.flatten())
    return np.asarray(jac)
# end doc example 3


def test_raise_Errors():
    with pytest.raises(CalculatorSetupError):
        HarmonicForceField(ref_atoms=ref_atoms, hessian_x=hessian_x,
                           get_q_from_x=lambda x: x)
    with pytest.raises(CalculatorSetupError):
        HarmonicForceField(ref_atoms=ref_atoms, hessian_x=hessian_x,
                           variable_orientation=True)
    with pytest.raises(CalculatorSetupError):
        HarmonicForceField(ref_atoms=ref_atoms, hessian_x=hessian_x,
                           cartesian=False)
    with pytest.raises(CalculationFailed):
        hff = HarmonicForceField(ref_atoms=ref_atoms, ref_energy=ref_energy,
                                 hessian_x=hessian_x,
                                 get_q_from_x=water_get_q_from_x,
                                 get_jacobian=lambda x: np.ones((3, 9)),
                                 cartesian=True, variable_orientation=True)
        calc = HarmonicCalculator(hff)
        setup_water(calc)


@pytest.mark.optimize()
def test_internals():
    parameters = {'ref_atoms': ref_atoms, 'ref_energy': ref_energy,
                  'hessian_x': hessian_x, 'get_q_from_x': water_get_q_from_x,
                  'get_jacobian': water_get_jacobian, 'cartesian': False}
    hff = HarmonicForceField(**parameters)  # calculation in internals
    calc = HarmonicCalculator(hff)
    atoms = setup_water(calc)  # distorted copy of ref_atoms
    run_optimize(atoms)        # recover original configuration
    assert_water_is_relaxed(atoms)

    parameters['cartesian'] = True  # calculation in Cartesian Coordinates
    hff = HarmonicForceField(**parameters)
    calc = HarmonicCalculator(hff)
    atoms = setup_water(calc)       # 'variable_orientation' not set to True!
    # but water has rotational degrees of freedom
    run_optimize(atoms)
    with pytest.raises(AssertionError):  # hence forces were incorrect
        assert_water_is_relaxed(atoms)   # original configuration not recovered

    parameters['variable_orientation'] = True
    hff = HarmonicForceField(**parameters)
    calc = HarmonicCalculator(hff)
    atoms = setup_water(calc)
    run_optimize(atoms)
    assert_water_is_relaxed(atoms)  # relaxation succeeded despite rotation


@pytest.mark.optimize()
def test_compatible_with_ase_vibrations():
    atoms = ref_atoms.copy()
    atoms.calc = EMT()
    run_optimize(atoms)
    opt_atoms = atoms.copy()
    opt_energy = atoms.get_potential_energy()
    vib = Vibrations(atoms, nfree=2)
    vib.run()
    energies = vib.get_energies()
    vib_data = vib.get_vibrations()
    hessian_2d = vib_data.get_hessian_2d()
    vib.clean()

    hff = HarmonicForceField(ref_atoms=opt_atoms, ref_energy=opt_energy,
                             hessian_x=hessian_2d)
    calc_harmonic = HarmonicCalculator(hff)
    atoms = ref_atoms.copy()
    atoms.calc = calc_harmonic
    vib = Vibrations(atoms, nfree=4, delta=1e-5)
    vib.run()
    assert np.allclose(energies, vib.get_energies())
    vib.clean()
    hff = HarmonicForceField(ref_atoms=ref_atoms, ref_energy=ref_energy,
                             hessian_x=hessian_2d,
                             get_q_from_x=water_get_q_from_x,
                             get_jacobian=water_get_jacobian, cartesian=True)
    calc_harmonic = HarmonicCalculator(hff)
    atoms = ref_atoms.copy()
    atoms.calc = calc_harmonic
    vib = Vibrations(atoms, nfree=4, delta=1e-5)
    vib.run()  # 3 transl and 3 rot are removed by internal coordinates
    assert_array_almost_equal(energies[-3:], vib.get_energies()[-3:], decimal=2)


def test_thermodynamic_integration():
    parameters = {'ref_atoms': ref_atoms, 'ref_energy': ref_energy,
                  'hessian_x': hessian_x, 'get_q_from_x': water_get_q_from_x,
                  'get_jacobian': water_get_jacobian, 'cartesian': True,
                  'variable_orientation': True}
    hff_1 = HarmonicForceField(**parameters)
    calc_harmonic_1 = HarmonicCalculator(hff_1)
    parameters['cartesian'] = False
    hff_0 = HarmonicForceField(**parameters)
    calc_harmonic_0 = HarmonicCalculator(hff_0)
    ediffs = {}  # collect energy difference for varying lambda coupling
    lambs = [0.00, 0.25, 0.50, 0.75, 1.00]  # integration grid
    for lamb in lambs:
        ediffs[lamb] = []
        calc_linearCombi = MixedCalculator(calc_harmonic_0, calc_harmonic_1,
                                           1 - lamb, lamb)
        atoms = ref_atoms.copy()
        atoms.calc = calc_linearCombi
        MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True)
        Stationary(atoms)
        ZeroRotation(atoms)
        with Andersen(atoms, 0.5 * fs, temperature_K=300, andersen_prob=0.05,
                      fixcm=False) as dyn:
            for _ in dyn.irun(50):  # should be much longer for production runs
                e0, e1 = calc_linearCombi.get_energy_contributions(atoms)
                ediffs[lamb].append(float(e1) - float(e0))
            ediffs[lamb] = np.mean(ediffs[lamb])
    dA = trapezoid([ediffs[lamb] for lamb in lambs], x=lambs)  # anharm. corr.
    assert -0.005 < dA < 0.005  # the MD run is to short for convergence
    if dA == 0.0:
        raise ValueError('there is most likely something wrong, but it could '
                         'also be sheer coincidence')



"""Build ASE's web-page."""

import os
import shutil
import subprocess
import sys
from pathlib import Path

cmds = """\
. /etc/bashrc
module load Tkinter
python3 -m venv venv
. venv/bin/activate
pip install -U pip
pip install "sphinx<6.0"  # search broken in sphinx-6
pip install sphinx-rtd-theme pillow
git clone http://gitlab.com/ase/ase.git
cd ase
pip install .
python setup.py sdist
cd doc
make
mv build/html ase-web-page"""


def build():
    root = Path('/tmp/ase-docs')
    if root.is_dir():
        sys.exit('Locked')
    root.mkdir()
    os.chdir(root)
    cmds2 = ' && '.join(line.split('#')[0] for line in cmds.splitlines())
    p = subprocess.run(cmds2, shell=True, check=False)
    if p.returncode == 0:
        status = 'ok'
    else:
        print('FAILED!', file=sys.stdout)
        status = 'error'
    f = root.with_name(f'ase-docs-{status}')
    if f.is_dir():
        shutil.rmtree(f)
    root.rename(f)
    return status


def build_all():
    assert build() == 'ok'
    tar = next(
        Path('/tmp/ase-docs-ok/ase/dist/').glob('ase-*.tar.gz'))
    webpage = Path('/tmp/ase-docs-ok/ase/doc/ase-web-page')
    home = Path.home() / 'web-pages'
    cmds = ' && '.join(
        [f'cp {tar} {webpage}',
         f'find {webpage} -name install.html | '
         f'xargs sed -i s/snapshot.tar.gz/{tar.name}/g',
         f'cd {webpage.parent}',
         'tar -czf ase-web-page.tar.gz ase-web-page',
         f'cp ase-web-page.tar.gz {home}'])
    subprocess.run(cmds, shell=True, check=True)


if __name__ == '__main__':
    build_all()



import numpy as np
import pytest

from ase.build import molecule


@pytest.mark.calculator_lite()
@pytest.mark.calculator('dftb')
def test_H2O(factory):
    atoms = molecule('H2O')
    atoms.calc = factory.calc(label='h2o')

    dipole = atoms.get_dipole_moment()

    assert np.linalg.norm(dipole) == pytest.approx(0.35212409930846317)



import numpy as np

from ase.build import graphene_nanoribbon


def test_get_angles():
    g = graphene_nanoribbon(3, 2, type="zigzag", vacuum=5)

    test_set = [[0, 1, x] for x in range(2, len(g))]

    manual_results = [g.get_angle(a1, a2, a3, mic=True)
                      for a1, a2, a3 in test_set]

    set_results = g.get_angles(test_set, mic=True)

    assert np.allclose(manual_results, set_results)



from pathlib import Path

import pytest

from ase.io import read
from ase.io.formats import UnknownFileTypeError


def mkfile(path, text):
    path = Path(path)
    path.write_text(text)
    return path


def test_no_such_file():
    fname = 'nosuchfile.traj'
    with pytest.raises(FileNotFoundError, match=fname):
        read(fname)


def test_empty_file():
    path = mkfile('empty.xyz', '')
    with pytest.raises(UnknownFileTypeError, match='Empty file'):
        read(path)


def test_bad_format():
    path = mkfile('strangefile._no_such_format', 'strange file contents')
    with pytest.raises(UnknownFileTypeError, match='_no_such_format'):
        read(path)
